"""Influx

This module provides methods for logging to an InfluxDB instance.
"""

import os
from dateutil.parser import parse
from influxdb import InfluxDBClient, SeriesHelper


def add_arguments(parser):

    parser.add_argument('--influxHost', dest="influx_host", type=str, nargs='?',
                        default=os.getenv("INFLUXDB_HOST", "terra-logging.ncsa.illinois.edu"),
                        help="InfluxDB URL for logging")
    parser.add_argument('--influxPort', dest="influx_port", type=int, nargs='?',
                        default= os.getenv("INFLUXDB_PORT", 8086),
                        help="InfluxDB port")
    parser.add_argument('--influxUser', dest="influx_user", type=str, nargs='?',
                        default=os.getenv("INFLUXDB_USER", "terra"),
                        help="InfluxDB username")
    parser.add_argument('--influxPass', dest="influx_pass", type=str, nargs='?',
                        default=os.getenv("INFLUXDB_PASSWORD", ''),
                        help="InfluxDB password")
    parser.add_argument('--influxDB', dest="influx_db", type=str, nargs='?',
                        default=os.getenv("INFLUXDB_DB", "extractor_db"),
                        help="InfluxDB database")


class Influx():

    def __init__(self, host, port, db, user, pass_):

        self.host = host
        self.port = port
        self.db = db
        self.user = user
        self.pass_ = pass_


    def log(self, extractorname, starttime, endtime, filecount, bytecount):

        f_completed_ts = int(parse(endtime).strftime('%s'))*1000000000
        f_duration = f_completed_ts - int(parse(starttime).strftime('%s'))*1000000000

        if self.pass_:
            client = InfluxDBClient(self.host, self.port, self.user,
                                    self.pass_, self.db)

            client.write_points([{
                "measurement": "file_processed",
                "time": f_completed_ts,
                "fields": {"value": f_duration}
            }], tags={"extractor": extractorname, "type": "duration"})
            client.write_points([{
                "measurement": "file_processed",
                "time": f_completed_ts,
                "fields": {"value": int(filecount)}
            }], tags={"extractor": extractorname, "type": "filecount"})
            client.write_points([{
                "measurement": "file_processed",
                "time": f_completed_ts,
                "fields": {"value": int(bytecount)}
            }], tags={"extractor": extractorname, "type": "bytes"})


    def error(self):
        # TODO: Allow sending critical error notification, e.g. email or Slack?
        pass


"""Users

This module provides access to user information
"""

import logging
import requests

logging.basicConfig(format='%(asctime)s %(message)s')


def get_dataset_username(host, key, dataset_id):
    """Looks up the name of the user associated with the dataset
    Args:
        host(str): the partial URI of the API path including protocol ('/api' portion and
                   after is not needed); assumes a terminating '/'
        key(str): access key for API use
        dataset_id(str): the id of the dataset belonging to the user to lookup
    Return:
        Returns the registered name of the found user. If the user is not found, None is
        returned. If a full name is available, that's returned. Otherwise the last name
        is returned and/or the first name (either both in that order, or one); a space
        separates the two names if they are concatenated and returned.
    Note:
        Any existing white space is kept intact for the name returned.
    Exceptions:
        HTTPError is thrown if a request fails
        ValueError ia thrown if the server returned data that is not JSON
    """
    # Initialize some variables
    user_id = None
    user_name = None

    # Get the dataset information
    url = "%sapi/datasets/%s?key=%s" % (host, dataset_id, key)
    result = requests.get(url)
    result.raise_for_status()

    # Get the author ID of the dataset
    ret = result.json()
    if 'authorId' in ret:
        user_id = ret['authorId']

    # Lookup the user information
    if not user_id is None:
        url = "%sapi/users/%s?key=%s" % (host, user_id, key)
        result = requests.get(url)
        result.raise_for_status()

        ret = result.json()
        if 'fullName' in ret:
            user_name = ret['fullName']
        else:
            if 'lastName' in ret:
                user_name = ret['lastName']
            if 'firstName' in ret:
                # pylint: disable=line-too-long
                user_name = ((user_name + ' ') if not user_name is None else '') + ret['firstName']

    return user_name

def find_user_name(host, secret_key, clowder_user, dataset_id=None):
    """Try to find the user in the clowder instance
    Args:
        host(str): the partial URI of the API path including protocol ('/api' portion and
                   after is not needed); assumes a terminating '/'
        secret_key(str): access key for API use
        clowder_user(str): the clowder username
        dataset_id(str): optional dataset identifier for looking up the user
    Return:
        Returns True if the user was found and False if not
    Exceptions:
        None
    """
    uris = ["%sapi/me?key=%s" % (host, secret_key),
            "%sapi/users?key=%s&limit=50000" % (host, secret_key)
           ]

    # Find additional places to look
    id_uris = []
    if not dataset_id is None:
        id_uris.append("%sapi/datasets/%s?key=%s" % (host, dataset_id, secret_key))
    for url in id_uris:
        try:
            result = requests.get(url)
            result.raise_for_status()

            # Get the author ID of the dataset
            ret = result.json()
            if 'authorId' in ret:
                user_id = ret['authorId']
                user_url = "%sapi/users/%s?key=%s" % (host, user_id, secret_key)
                uris.insert(0, user_url)
        # pylint: disable=broad-except
        except Exception:
            pass
        # pylint: enable=broad-except

    # Now look through all the places to look
    for url in uris:
        try:
            result = requests.get(url)
            result.raise_for_status()

            ret = result.json()
            if not isinstance(ret, list):
                ret = [ret]

            for user in ret:
                if ("email" in user) and (user["email"] == clowder_user):
                    return True
        # pylint: disable=broad-except
        except Exception:
            pass
        # pylint: enable=broad-except

    # We didn't find the user
    return False



"""LemnaTec

This module is used to standardize LemnaTec metadata.

Given a JSON object containing the original LemnaTec metadata, convert
field names and values to standardized formats required for TERRA-REF.

The "cleaned" metadata will be posted to the Clowder metadata endpoint and
used downstream by extractors.

TODO:
    SENSOR_IRRIGATION                                                                                   
    SENSOR_LIGHTNING
    SENSOR_ENVIRONMENTAL_LOGGER
    SENSOR_WEATHER
"""

import argparse
import json
import csv
import logging
import os
import re
import pytz, datetime

import terrautils.betydb
from terrautils.sensors import Sensors
from terrautils.spatial import calculate_gps_bounds, calculate_centroid, tuples_to_geojson


STATION_NAME = "ua-mac"

scan_programs = {}

# Official sensor names
PLATFORM_SCANALYZER = "scanalyzer"
SENSOR_CO2 = "co2Sensor"
SENSOR_CROP_CIRCLE = "cropCircle"
SENSOR_ENVIRONMENTAL_LOGGER = "EnvironmentLogger"
SENSOR_FLIR = "flirIrCamera"
SENSOR_IRRIGATION = "irrigation"
SENSOR_LIGHTNING = "lightning"
SENSOR_NDVI = "ndviSensor"
SENSOR_PAR = "parSensor"
SENSOR_PRI = "priSensor"
SENSOR_PS2_TOP = "ps2Top"
SENSOR_SCANNER_3D_TOP = "scanner3DTop"
SENSOR_STEREO_TOP = "stereoTop"
SENSOR_SWIR = "SWIR"
SENSOR_VNIR = "VNIR"
SENSOR_WEATHER = "weather"

SENSOR_METADATA_CACHE = os.environ.get('SENSOR_METADATA_CACHE', '/home/extractor/sites/ua-mac/sensor-metadata')

logging.basicConfig()
logger = logging.getLogger("terrautils.metadata.lemnatac")



# SHARED -------------------------------------
def clean(metadata, sensorId, filepath="", fixed=False):
    """ 
    Given a LemnaTec metadata.json object, produces the "cleaned" metadata that 
    will be put in the Clowder jsonld endpoint.
    """
    
    orig_lem_md = metadata['lemnatec_measurement_metadata']
    
    cleaned_md = {}
    cleaned_md["gantry_variable_metadata"] = _standardize_gantry_system_variable_metadata(orig_lem_md, filepath)
    cleaned_md["gantry_fixed_metadata"]    = _get_sensor_fixed_metadata_json(PLATFORM_SCANALYZER)
    cleaned_md["sensor_fixed_metadata"]    = _get_sensor_fixed_metadata_json(sensorId)
    cleaned_md["sensor_variable_metadata"] = _standardize_sensor_variable_metadata(sensorId, orig_lem_md, 
                                                    cleaned_md["gantry_variable_metadata"], filepath)
    date = cleaned_md["gantry_variable_metadata"]["date"]
    cleaned_md["terraref_cleaned_metadata"] = True
    
    # calculate_gps_bounds requires the fixed metadata FOV
    full_md = cleaned_md.copy()
    query_date = _get_date_from_raw_metadata(orig_lem_md)
    fixed_md = _get_sensor_fixed_metadata(sensorId, query_date)
    full_md["sensor_fixed_metadata"] = fixed_md
    if fixed:
        cleaned_md["sensor_fixed_metadata"] = fixed_md

    cleaned_md["experiment_metadata"] = _get_experiment_metadata(date, sensorId)
    cleaned_md["site_metadata"] = _get_sites(full_md, date, sensorId)
    cleaned_md["spatial_metadata"] = _get_spatial_metadata(full_md, sensorId)
    return cleaned_md


# PRIVATE -------------------------------------
def _get_spatial_metadata(cleaned_md, sensorId):
    gps_bounds = calculate_gps_bounds(cleaned_md, sensorId) 

    spatial_metadata = {}
    for label, bounds in gps_bounds.iteritems():
        spatial_metadata[label] = {}
        spatial_metadata[label]["bounding_box"] = tuples_to_geojson(bounds)
        spatial_metadata[label]["centroid"] = calculate_centroid(bounds)
        
    return spatial_metadata


def _get_sites(cleaned_md, date, sensorId):
    """
    Returns the site name and URL for all sites associated with the centroid.
    """
    gps_bounds = calculate_gps_bounds(cleaned_md, sensorId)

    sites = {}
    for label, bounds in gps_bounds.iteritems():
        centroid = calculate_centroid(bounds)
        bety_sites = terrautils.betydb.get_sites_by_latlon(centroid, date)
        for bety_site in bety_sites:
            sites['id'] = str(bety_site["id"])
            sites['notes'] = "plot that contains the image centroid"
            sites["sitename"] = bety_site["sitename"]
            if "view_url" in bety_site:
                sites["url"] = bety_site["view_url"]
            else:
                sites["url"] = ""

    return sites


def _get_experiment_metadata(date, sensorId): 
    sensors = Sensors(base="", station=STATION_NAME, sensor=sensorId)
    exps = sensors.get_experiment(date)
    
    experiment_md = []
    for exp in exps:
        curr_exp = {}
        curr_exp["name"] = exp["name"]
        curr_exp["start_date"] = exp["start_date"]
        curr_exp["end_date"] = exp["end_date"]
        curr_exp["url"] = exp["view_url"]
        experiment_md.append(curr_exp)

    return experiment_md    


def _standardize_gantry_system_fixed_metadata(orig):
    """
    Returns an object containing the URL for the Scanalyzer fixed metadata
    """
    properties = {}
    properties["url"] = _get_sensor_fixed_metadata_json(PLATFORM_SCANALYZER)
    return properties


def _get_sensor_fixed_metadata_json(sensorId):
    """
    Assumes that the sensor fixed metadata stored in Clowder is authoritative
    Ignore the fixed metadata in the JSON object and return the fixed metadata URL in Clowder.
    """
    # TODO: Compare to known fixed metadata structure
    # TODO; We only need this one -- duplicate method in metadata.py
    
    # Get the json path for the sensor by identifier
    sensors = Sensors(base="", station=STATION_NAME, sensor=sensorId)
    jsonpath = sensors.get_fixed_jsonpath_for_sensor()

    return {
        "url": "https://github.com/terraref/sensor-metadata/blob/master/" + jsonpath
    }


def _get_sensor_fixed_metadata(sensorId, query_date):
    # Get the json path for the sensor by identifier
    sensors = Sensors(base="", station=STATION_NAME, sensor=sensorId)
    jsonpath = sensors.get_fixed_jsonpath_for_sensor()

    sensor_file = SENSOR_METADATA_CACHE + jsonpath
    if os.path.exists(sensor_file):
        with open(sensor_file, 'r') as sf:
            md_json = json.load(sf)
            if type(md_json) == list:
                if type(md_json[0] == dict):
                    current_metadata = find_json_for_date(query_date, md_json)
            return current_metadata
    else:
        # TODO: What should happen here?
        return None


def _standardize_gantry_system_variable_metadata(lem_md, filepath=""):
    """
    Standardize the gantry variable metadata.  Note, the original LemnaTec metadata
    changes keys over time (e.g., time=Time=timestamp=TimeStamp).  The prop_map
    contains all keys encountered in the gantry_system_variable_metadata over time,
    although many of these are never used in the final cleaned metadata.
    """   
    
    prop_map = {
        'time': {
            'standardized': ['time']
        },
        'Time': {
            'standardized': ['time']
        },        
        'timestamp': {
            'standardized': ['time'] 
        }, 
        'Timestamp': {
            'standardized': ['time'] 
        },         
        'position x [m]': {
            'standardized': ['position_m', 'x']
        },
        'Position x [m]': {
            'standardized': ['position_m', 'x']
        },        
        'position y [m]': {
            'standardized': ['position_m', 'y']
        },
        'Position y [m]': {
            'standardized': ['position_m', 'y']
        },         
        'position z [m]': {
            'standardized': ['position_m', 'z']   
        },
        'Position z [m]': {
            'standardized': ['position_m', 'z']
        },          
        'speed x [m/s]': {
            'standardized': ['speed_m/s', 'x']
        },
        'speed y [m/s]': {
            'standardized': ['speed_m/s', 'y']
        },  
        'speed z [m/s]': {
            'standardized': ['speed_m/s', 'z']
        },
        'Velocity x [m/s]': {
            'standardized': ['velocity_m/s', 'x']
        },          
        'Velocity y [m/s]': {
            'standardized': ['velocity_m/s', 'y']
        },         
        'Velocity z [m/s]': {
            'standardized': ['velocity_m/s', 'z']
        },         
        'scanDistance [m]': {
            'standardized': ['scan_distance_m']
        },
        'scanDistanceInM [m]': {
            'standardized': ['scan_distance_m']
        },        
        'scanSpeed [m/s]': {
            'standardized': ['scan_speed_m/s']
        }, 
        'scanSpeedInMPerS [m/s]': {
            'standardized': ['scan_speed_m/s']
        },         
        'scanMode': {
            'standardized': ['scan_mode']
        },         
        'camera box light 1 is on': {
            'standardized': ['camera_box_light_1_on']
        }, 
        'Camnera box light 1 is on': {
            'standardized': ['camera_box_light_1_on']
        },          
        'camera box light 2 is on': {
            'standardized': ['camera_box_light_2_on']
        },  
        'Camnera box light 2 is on': {
            'standardized': ['camera_box_light_2_on']
        },        
        'camera box light 3 is on': {
            'standardized': ['camera_box_light_3_on']
        },  
        'Camnera box light 3 is on': {
            'standardized': ['camera_box_light_3_on']
        },        
        'camera box light 4 is on': {
            'standardized': ['camera_box_light_4_on']
        },  
        'Camnera box light 4 is on': {
            'standardized': ['camera_box_light_4_on']
        },         
        'Script copy path on FTP server': {
            'standardized': ['script_path_ftp_server']
        },   
        'Script path on local disk': {
            'standardized': ['script_path_on_disk']
        },           
        'sensor setting file path': {
            'standardized': ['sensor_setting_file_path']
        },  
        # This is used in the calculation of the point cloud origin
        'scanIsInPositiveDirection': {
            'standardized': ['scan_direction_is_positive'],
            'default' : "False"
        }, 
        'scanDirectionIsPositive': {
            'standardized': ['scan_direction_is_positive'],
            'default' : "False"
        },
        'PLC control not available': {
            'standardized': ['error']
        },
        # Found on co2Sensor
        'x end pos [m]': {
            'standardized': ['end_position_m', 'x']
        },     
        'x set velocity [m/s]': {
            'standardized': ['velocity_m/s', 'x']
        }, 
        'x set acceleration [m/s^2]': {
            'standardized': ['acceleration_m/s^2', 'x']
        },
        'x set deceleration [m/s^2]': {
            'standardized': ['deceleration_m/s^2', 'x']
        },          
        # Found on cropCircle
        'y end pos [m]': {
            'standardized': ['end_position_m', 'y']
        },  
        'Y end pos [m]': {
            'standardized': ['end_position_m', 'y']
        },          
        'y set velocity [m/s]': {
            'standardized': ['velocity_m/s', 'y']
        },
        'Y set velocity [m/s]': {
            'standardized': ['velocity_m/s', 'y']
        },         
        'y set acceleration [m/s^2]': {
            'standardized': ['acceleration_m/s^2', 'y']
        },
        'Y set acceleration [m/s^2]': {
            'standardized': ['acceleration_m/s^2', 'y']
        },        
        'y set deceleration [m/s^2]': {
            'standardized': ['deceleration_m/s^2', 'y']
        },
        'y set decceleration [m/s^2]': {
            'standardized': ['deceleration_m/s^2', 'y']
        },        
        'Y set decceleration [m/s^2]': {
            'standardized': ['deceleration_m/s^2', 'y']
        },
        # FAT Tests?
        'Measurement purpose [FAT test]' : {
            'standardized': ['ignored']
        },
        'Measurement target [Test object sandbox]' : {
            'standardized': ['ignored']
        },
        'fat measurement [comparison color sensors to hyperspec above green target]' : {
            'standardized': ['ignored']
        },
        'repeats [1 of 1]' : {
            'standardized': ['ignored']
        },        
        'repeats [1 of 3]' : {
            'standardized': ['ignored']
        },
        'repeats [2 of 3]' : {
            'standardized': ['ignored']
        },
        'repeats [3 of 3]' : {
            'standardized': ['ignored']
        },
        'fat measurement [test chart (resting)]' : {
            'standardized': ['ignored'],
            'default': "test chart(resting)"
        },
        'fat measurement [black body (moving)]' : {
            'standardized': ['ignored']
        },
        'fat measurement [black body (resting)]' : {
            'standardized': ['ignored']
        },
        'only small set of meta data available' : {
            'standardized': ['ignored']
        },
        'Only small set of meta data available' : {
            'standardized': ['ignored']
        },        
        'fat measurement [PS2 on fluo target]': {
            'standardized': ['ignored']
        },
        'fat measurement [3d scan of test target (different directions and different speeds)]': {
            'standardized': ['ignored']
        }
        
    }
    
    
    orig = lem_md['gantry_system_variable_metadata'] 
    properties = _standardize_with_validation("gantry_system_variable_metadata", orig, prop_map, [], filepath)  

    # Set default scan_direction_is_positive
    if 'scan_direction_is_positive' not in properties:
        if 'position_m' in properties:
            if properties['position_m']['y'] == 0:
                properties['scan_direction_is_positive'] = 'True'
            else:
                properties['scan_direction_is_positive'] = 'False'
                    
    # Standardize time field value
    if 'time' in properties:
        datetime_local = _standardize_time(properties['time'], "%m/%d/%Y %H:%M:%S", "US/Arizona")
        properties["datetime"] = datetime_local.isoformat()
        properties["date"] = datetime_local.date().isoformat()
        
    # 
    read_scan_program_map()
    if 'script_path_on_disk' in properties:
        # e.g. "C:\\LemnaTec\\StoredScripts\\3D_Scan_Field_SouthStart_033MperS.cs"
        script_path = properties['script_path_on_disk']
        scan_program = scan_programs.get(script_path, "unknown_program")
        match = re.search('^.*\\\\(.*?)\\.cs', script_path)   
        if len(match.groups()) == 1:
            script_name_caps = match.group(1)
            script_name = script_name_caps.lower()
            script_name = re.sub(" ", "_", script_name)
            properties['script_name'] = script_name
            if isinstance(scan_program, list):
                properties['fullfield_eligible'] = scan_program["fullfield_eligible"]
            else:
                properties['fullfield_eligible'] = "True"

            # Also retain unique script hash to differentiate if same scan run multiple times in a day
            if 'script_path_ftp_server' in properties:
                # e.g. "ftp://10.160.21.2//gantry_data/LemnaTec/ScriptBackup/3D_Scan_Field_SouthStart_033MperS_6d2cf837-5107-4a67-87f3-cc7b65551931.cs"
                script_hash = properties['script_path_ftp_server']
                # e.g. "6d2cf837-5107-4a67-87f3-cc7b65551931"
                script_hash = script_hash[script_hash.find(script_name_caps)+len(script_name_caps)+1:-3]
                properties['script_hash'] = script_hash
        
    # Limit output to the following fields for now
    output_fields = [
        "datetime", "date", "position_m", "scan_direction_is_positive", "script_path_on_disk", "script_name", "script_hash", "fullfield_eligible", "error"
    ]

    return _get_dict_subset(properties, output_fields)


def _get_date_from_raw_metadata(md):
    default = "2012-01-01"
    if "gantry_system_variable_metadata" in md:
        if "time" in md["gantry_system_variable_metadata"]:
            t = md["gantry_system_variable_metadata"]["time"].split(" ")[0]
            dd = t.split("/")[1]
            mm = t.split("/")[0]
            yy = t.split("/")[2]
            return yy+"-"+mm+"-"+dd

        else:
            return default
    else:
        return default


def _standardize_sensor_variable_metadata(sensor, orig_lem_md, corrected_gantry_variable_md, filepath=""):
    """
    Standardize the sensor variable metadata
    
    TODO:
        SENSOR_IRRIGATION                                                                                   
        SENSOR_LIGHTNING
        SENSOR_ENVIRONMENTAL_LOGGER
        SENSOR_WEATHER
    """

    query_date = _get_date_from_raw_metadata(orig_lem_md)
    sensor_variable_metadata = orig_lem_md['sensor_variable_metadata'] 
    sensor_fixed_metadata = _get_sensor_fixed_metadata(sensor, query_date)

    if sensor == SENSOR_CO2:
        properties = _co2_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_CROP_CIRCLE:
        properties = _cropCircle_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_ENVIRONMENTAL_LOGGER:
        properties = _xxx_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_FLIR:
        properties = _flir_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_NDVI:
        properties = _ndvi_standardize(sensor_variable_metadata, filepath)        
    elif sensor == SENSOR_PAR:
        properties = _par_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_PRI:
        properties = _pri_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_PS2_TOP:
        properties = _ps2_standardize(sensor_variable_metadata, filepath)        
    elif sensor == SENSOR_SCANNER_3D_TOP:
        properties = _scanner3d_standardize(sensor_variable_metadata, sensor_fixed_metadata, 
                                                corrected_gantry_variable_md, filepath)
    elif sensor == SENSOR_STEREO_TOP:
        properties = _stereoTop_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_SWIR:
        properties = _swir_standardize(sensor_variable_metadata, filepath)
    elif sensor == SENSOR_VNIR:
        properties = _vnir_standardize(sensor_variable_metadata, filepath)

    return properties


def _standardize_with_validation(name, orig, property_map, required_fields=[], filepath=""):
    """
    Use the property_map to standardize the original metadata, using default values where specified.
    Log any warnings related to missing fields and errors for expected/required fields.
    """
    standardized = {}

    # Step through the fields in the original metadata and convert to the standardized form.
    # Warn if we don't have a mapping.
    for key in orig:
        if key in property_map:
            _set_nested_value(standardized, property_map[key]['standardized'], orig[key])
        else:
            logger.warning("Encountered field \"%s\", missing from map in %s (%s)"  % (key, name, filepath))

    # Step through the keys of the mapping, set default values where appropriate and
    # report an error if a required field is missing.
    for key in property_map:
        if key in required_fields and not _nested_contains(standardized, property_map[key]['standardized']):
            if key in property_map and 'default' in property_map[key]:
                logger.debug("Setting default value %s for key \"%s\"" % ( property_map[key]['default'],
                                                                           property_map[key]['standardized']))

                _set_nested_value(standardized, property_map[key]['standardized'], property_map[key]['default'])
            else:
                logger.error("missing required field \"%s\" in %s" % (property_map[key]['standardized'], name))
    return standardized


def _nested_contains(dic, keys):
    """
    Returns true if the keys exist
    """
    for key in keys[:-1]:
        if key in dic:
            dic = dic.get(key)
        else:
            return False

    return (keys[-1] in dic)


def _set_nested_value(dic, keys, value):
    """
    Given a set of keys as an array, sets a nested dictionary value. This is used to convert properties
    such as "position_x" or "position_y" to "position[x]" and "position[y]"
    """
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def _standardize_time(timestr, timeformat, timezone):
    """
    Given the timestring, format, and timezone string, return the date/time in ISO 8601 format.
    """
    tz = pytz.timezone (timezone)
    time = datetime.datetime.strptime (timestr, timeformat)
    local_dt = tz.localize(time, is_dst=None)
    #utc_dt = local_dt.astimezone (pytz.utc)
    return (local_dt)


def _get_dict_subset(dic, keys):
    """
    Return a subset of a dictionary containing only the specified keys.
    """
    return dict((k, dic[k]) for k in keys if k in dic)


# SENSOR-SPECIFIC -------------------------------------
def _cropCircle_standardize(data, filepath=""):

    prop_map = {
        "current setting rotate flip type": {
            "standardized": ["rotate_flip_type"]
        },
        "current setting crosshairs": {
            "standardized": ["crosshairs"]
        }      
    }

    properties = _standardize_with_validation(SENSOR_CROP_CIRCLE, data, prop_map, [], filepath="")   
    return properties


def _flir_standardize(data, filepath=""):

    prop_map = {
        "current setting AutoFocus": {
            "standardized": ["autofocus"]
        },
        "current setting Manual focal length [cm]": {
            "standardized": ["manual_focal_length_cm"]
        },
        # 2016 data
        "current setting Manual focal length": {
            "standardized": ["manual_focal_length_cm"]
        },        
        "current setting ImageAdjustMode": {
            "standardized": ["image_adjust_mode"]
        },            
        "camera info": {
            "standardized": ["camera_info"]
        },
        "focus distance [m]": {
            "standardized": ["focus_distance_m"]
        },
        "lens temperature [K]": {
            "standardized": ["lens_temperature_K"]
        },
        "shutter temperature [K]": {
            "standardized": ["shutter_temperature_K"]
        }, 
        "front temperature [K]": {
            "standardized": ["front_temperature_K"]
        }               
    }
    
    properties = _standardize_with_validation(SENSOR_FLIR, data, prop_map, [], filepath)    
    return properties    
    
    
def _ps2_standardize(data, filepath=""):

    prop_map = {
        "current setting rotate flip type" : {
            "standardized": ["rotate_flip_type"]
        },
        "current setting crosshairs" : {
            "standardized": ["crosshairs"]
        },
        "current setting exposure" : {
            "standardized": ["exposure"]
        },
        "current setting gain" : {
            "standardized": ["gain"]
        },        
        "current setting gamma" : {
            "standardized": ["gamma"]
        },  
        "current setting ledcurrent" : {
            "standardized": ["led_current"]
        }           
    }
    properties = _standardize_with_validation(SENSOR_PS2_TOP, data, prop_map, [], )   
    return properties    


def _stereoTop_standardize(data, filepath=""):
    prop_map = {
        "Rotate flip type - left" : {
            "standardized": ["rotate_flip_type", "left"]
        },
        "Rotate flip type - right" : {
            "standardized": ["rotate_flip_type", "right"]
        },  
        "rotate flip type - left" : {
            "standardized": ["rotate_flip_type", "left"]
        },
        "rotate flip type - right" : {
            "standardized": ["rotate_flip_type", "right"]
        },          
        "Crosshairs - left" : {
            "standardized": ["crosshairs", "left"]
        },
        "Crosshairs - right" : {
            "standardized": ["crosshairs", "right"]
        },    
        "crosshairs - left" : {
            "standardized": ["crosshairs", "left"]
        },
        "crosshairs - right" : {
            "standardized": ["crosshairs", "right"]
        },         
        "exposure - left" : {
            "standardized": ["exposure", "left"]
        },
        "exposure - right" : {
            "standardized": ["exposure", "right"]
        },        
        "autoexposure - left" : {
            "standardized": ["autoexposure", "left"]
        },
        "autoexposure - right" : {
            "standardized": ["autoexposure", "right"]
        },        
        "gain - left" : {
            "standardized": ["gain", "left"]
        },  
        "gain - right" : {
            "standardized": ["gain", "right"]
        },         
        "autogain - left" : {
            "standardized": ["autogain", "left"]
        },  
        "autogain - right" : {
            "standardized": ["autogain", "right"]
        },          
        "gamma - left" : {
            "standardized": ["gamma", "left"]
        }, 
        "gamma - right" : {
            "standardized": ["gamma", "right"]
        },         
        "rwhitebalanceratio - left" : {
            "standardized": ["rwhitebalanceratio", "left"]
        }, 
        "rwhitebalanceratio - right" : {
            "standardized": ["rwhitebalanceratio", "right"]
        },           
        "bwhitebalanceratio - left" : {
            "standardized": ["bwhitebalanceratio", "left"]
        },  
        "bwhitebalanceratio - right" : {
            "standardized": ["bwhitebalanceratio", "right"]
        },         
        "height left image [pixel]" : {
            "standardized": ["height_image_pixels", "left"]
        },  
        "width left image [pixel]" : {
            "standardized": ["width_image_pixels", "left"]
        },       
        "image format left image" : {
            "standardized": ["image_format", "left"]
        },       
        "height right image [pixel]" : {
            "standardized": ["height_image_pixels", "right"]
        },  
        "width right image [pixel]" : {
            "standardized": ["width_image_pixels", "right"]
        },       
        "image format right image" : {
            "standardized": ["image_format", "right"]
        }, 
    }
    
    properties = _standardize_with_validation(SENSOR_STEREO_TOP, data, prop_map, [], filepath)  
    return properties     
    
    
def _swir_standardize(data, filepath=""):
    """
    Standardize SWIR metadata (same format as VNIR)
    """
    return _vnir_standardize(data, filepath, SENSOR_SWIR)
    
    
def _vnir_standardize(data, filepath="", name=SENSOR_VNIR):
    """
    Standardize VNIR metadata (same format as VNIR)
    """    
    prop_map = {
        "current setting frameperiod": {
            "standardized": ["frame_period"]
        },
        "current setting userotatingmirror": {
            "standardized": ["use_rotating_mirror"]
        },
        "current setting useexternaltrigger": {
            "standardized": ["use_external_trigger"]
        },       
        "current setting exposure": {
            "standardized": ["exposure"]
        },
        "current setting createdatacube": {
            "standardized": ["create_data_cube"]
        },      
        "current setting speed": {
            "standardized": ["speed"]
        }, 
        "current setting constmirrorpos": {
            "standardized": ["const_mirror_position"]
        },  
        "current setting startpos": {
            "standardized": ["start_position"]
        }, 
        "current setting stoppos": {
            "standardized": ["stop_position"]
        }
    }
    
    properties = _standardize_with_validation(name, data, prop_map, [], filepath)  
    return properties


def _co2_standardize(data, filepath=""):
    """
    Placeholder only, no variable metadata
    See /data/terraref/sites/ua-mac/raw_data/co2Sensor/2017-06-27/2017-06-27__13-32-28-129/17e118dc-20fb-4b59-9e59-9ee9840f302a_metadata.json
    """
    return {}


def _pri_standardize(data, filepath=""):
    """
    Placeholder only, no variable metadata
    See  /data/terraref/sites/ua-mac/raw_data/priSensor/2017-06-27/2017-06-27__13-32-28-039/baa2813f-0634-45df-8e9b-4a978fa93f86_metadata.json
    """
    return {}


def _ndvi_standardize(data, filepath=""):
    """
    Placeholder only, no variable metadata
    See   /data/terraref/sites/ua-mac/raw_data/ndviSensor//2017-03-13/2017-03-13__13-56-55-559/2d4ae02b-3475-42a8-bb73-fe972f256aaf_metadata.json    """
    return {}


def _scanner3d_standardize(data, fixed_md, corrected_gantry_variable_md, filepath=""):
    prop_map = {
        "current setting Exposure [microS]": {
            "standardized": ["exposure_microS"]
        },
        # 2016 data
        "current setting Exposure": {
            "standardized": ["exposure_microS"]
        },
        "current setting Calculate 3D files": {
            "standardized": ["calculate_3d_files"]
        },
        "current setting Laser detection threshold": {
            "standardized": ["laser_detection_threshold"]
        },
        "current setting Scanlines per output file": {
            "standardized": ["scanlines_per_output_file"]
        },
        "current setting Scan direction (automatically set at runtime)": {
            "standardized": ["scan_direction"]
        },
        "current setting Scan distance (automatically set at runtime) [mm]": {
            "standardized": ["scan_distance_mm"]
        },
        "current setting Scan speed (automatically set at runtime) [microMeter/s]": {
            "standardized": ["scan_speed_microMeter/s"]
        },
        "current setting Scan speed (automatically set at runtime)": {
            "standardized": ["scan_speed_microMeter/s"]
        },
        "current setting Scan distance (automatically set at runtime)": {
            "standardized": ["scan_distance_mm"]
        },
    }

    properties = _standardize_with_validation(SENSOR_SCANNER_3D_TOP, data, prop_map, [], filepath)
    properties["point_cloud_origin_m"] = _calculatePointCloudOrigin(data, fixed_md, corrected_gantry_variable_md)
    return properties


def _calculatePointCloudOrigin(scanner3d, fixed_md, gantvar_md):
    '''
        Calculate the origin of the point cloud. 
        Per https://github.com/terraref/reference-data/issues/44
            * The origin of point cloud in Z direction is the subtraction of 3445mm from the gantry position during the scan.
            * In X direction, is +82mm to the north from the center of the 3D scanner.
            * If the scan is in positive direction, the origin of ply files in Y direction is +3450mm in gantry coordinate system, 
            * If the scan is done in negative direction is +25711mm in gantry coordinate system.
        The origin is calculated based on the position of the west scanner. 
        So, any further misalignment correction should be applied to the east ply files.
        If Y is zero, then scan_direction_is_positive
        plc_control_not_available means there is no logation data (position_m)

        gantvar_md = corrected gantry variable metadata
    '''
    
    point_cloud_origin = {"east": {}, "west": {}}
    if (not 'plc_control_not_available' in gantvar_md
            and 'position_m' in gantvar_md
            and 'scan_direction_is_positive' in gantvar_md
            and 'scanner_west_location_in_camera_box_m' in fixed_md):

        cambox_e = fixed_md['scanner_east_location_in_camera_box_m']
        cambox_w = fixed_md['scanner_west_location_in_camera_box_m']

        xv = float(gantvar_md["position_m"]["x"]) + 0.082
        point_cloud_origin['east']["x"] = xv + float(cambox_e['x'])
        point_cloud_origin['west']["x"] = xv + float(cambox_w['x'])

        zv = float(gantvar_md['position_m']['z']) - 3.445
        point_cloud_origin['east']["z"] = zv + float(cambox_e['z'])
        point_cloud_origin['west']["z"] = zv + float(cambox_w['z'])

        if (gantvar_md["scan_direction_is_positive"] == "True"):
            yv = float(gantvar_md['position_m']['y']) + 3.450
        else:
            yv = float(gantvar_md['position_m']['y']) + 25.711
        point_cloud_origin['east']["y"] = yv + float(cambox_e['y'])*2
        point_cloud_origin['west']["y"] = yv + float(cambox_w['y'])*2

    else:
        logger.error("Cannot calculate point cloud origin -- missing gantry position information")

    return point_cloud_origin


def read_scan_program_map():

    if not scan_programs:
        scan_path = os.path.join(os.path.dirname(__file__), "data/scan_programs.csv")
        with open(scan_path, 'rb') as file:
            reader = csv.DictReader(file)
            for row in reader:
                scan_programs[row["program_name"]] = {
                    "fullfield_eligible": "True" if row["fullfield_eligible"] == "Y" else "False"
                }


def get_content_for_date(query_date, content_json):
    query_datetime = datetime.datetime.strptime(query_date, '%Y-%m-%d')
    current_match = None
    if 'start_dates' in content_json:
        print("we have start dates")
        start_dates_raw = content_json['start_dates'].keys()
        start_dates = []
        for each in start_dates_raw:
            print('each', each)
            current = datetime.datetime.strptime(each, '%Y-%m-%d')
            start_dates.append(current)

        start_dates.sort()
        for sd in start_dates:
            if query_datetime < sd:
                current_match = sd
                print('new current match', current_match)
        match_string = str(current_match)
        match_string = match_string[0:match_string.index(' ')]
        return content_json['start_dates'][match_string]


def find_json_for_date(query_date, json_list):
    if len(json_list) == 1:
        return json_list[0]
    query_date = datetime.datetime.strptime(query_date, '%Y-%m-%d')
    best_match = None
    best_match_date = None

    for each in json_list:
        metadata_date = datetime.datetime.strptime(each['start_date'], '%Y-%m-%d')

        if best_match is None:
            if query_date >= metadata_date:
                best_match = each
                best_match_date = datetime.datetime.strptime(each['start_date'], '%Y-%m-%d')
        else:
            if query_date > metadata_date > best_match_date:
                best_match = each
                best_match_date = datetime.datetime.strptime(each['start_date'], '%Y-%m-%d')
    return best_match



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=str, help="Path to metadata json")
    parser.add_argument("sensor", type=str, help="Sensor name")
    parser.add_argument("--output", help="Print output", action="store_true")
    parser.add_argument("--debug", help="Print debug output", action="store_true")
    args = parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)
        
    logger.debug("Processing %s" % args.path)

    with open(args.path) as file:
        json_data = json.load(file)

    clean_md = [{"content":  clean(json_data, args.sensor)}]
    
    if args.output:
        print(json.dumps(clean_md, indent=4, sort_keys=False))



"""Formats

This module handles creation of output files in GeoTIF, netCDF, and other image formats.
"""

import numpy
import subprocess
import os
from osgeo import gdal, osr
from netCDF4 import Dataset
from matplotlib import cm, pyplot as plt
from PIL import Image


def compress_geotiff(input_file):
    temp_out = input_file.replace(".tif", "_compress.tif")
    subprocess.call(["gdal_translate", "-co", "COMPRESS=LZW", input_file, temp_out])
    if os.path.isfile(temp_out):
        os.remove(input_file)
        os.rename(temp_out, input_file)


def create_geotiff(pixels, gps_bounds, out_path, nodata=-99, asfloat=False, extractor_info=None, system_md=None, extra_metadata=None, compress=False):
    """Generate output GeoTIFF file given a numpy pixel array and GPS boundary.

        Keyword arguments:
        pixels -- numpy array of pixel values.
                    if 2-dimensional array, a single-band GeoTIFF will be created.
                    if 3-dimensional array, a band will be created for each Z dimension.
        gps_bounds -- tuple of GeoTIFF coordinates as ( lat (y) min, lat (y) max,
                                                        long (x) min, long (x) max)
        out_path -- path to GeoTIFF to be created
        nodata -- NoDataValue to be assigned to raster bands; set to None to ignore
        float -- whether to use GDT_Float32 data type instead of GDT_Byte (e.g. for decimal numbers)
        extractor_info -- details about extractor if applicable
        system_md -- cleaned TERRA-REF metadata
        extra_metadata -- any metadata to be embedded in geotiff; supersedes extractor_info and system_md
    """
    dimensions = numpy.shape(pixels)
    if len(dimensions) == 2:
        nrows, ncols = dimensions
        channels = 1
    else:
        nrows, ncols, channels = dimensions

    geotransform = (
        gps_bounds[2], # upper-left x
        (gps_bounds[3] - gps_bounds[2])/float(ncols), # W-E pixel resolution
        0, # rotation (0 = North is up)
        gps_bounds[1], # upper-left y
        0, # rotation (0 = North is up)
        -((gps_bounds[1] - gps_bounds[0])/float(nrows)) # N-S pixel resolution
    )

    # Create output GeoTIFF and set coordinates & projection
    if asfloat:
        dtype = gdal.GDT_Float32
    else:
        dtype = gdal.GDT_Byte

    if compress:
        output_raster = gdal.GetDriverByName('GTiff') \
            .Create(out_path, ncols, nrows, channels, dtype, ['COMPRESS=LZW'])
    else:
        output_raster = gdal.GetDriverByName('GTiff') \
            .Create(out_path, ncols, nrows, channels, dtype)

    output_raster.SetGeoTransform(geotransform)
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326) # google mercator
    output_raster.SetProjection( srs.ExportToWkt() )

    if not extra_metadata:
        extra_metadata = prepare_metadata_for_geotiff(extractor_info, system_md)

    output_raster.SetMetadata(extra_metadata)

    if channels == 3:
        # typically 3 channels = RGB channels
        # TODO: Something wonky w/ uint8s --> ending up w/ lots of gaps in data (white pixels)
        output_raster.GetRasterBand(1).WriteArray(pixels[:,:,0].astype('uint8'))
        output_raster.GetRasterBand(1).SetColorInterpretation(gdal.GCI_RedBand)
        output_raster.GetRasterBand(1).FlushCache()
        if nodata:
            output_raster.GetRasterBand(1).SetNoDataValue(nodata)

        output_raster.GetRasterBand(2).WriteArray(pixels[:,:,1].astype('uint8'))
        output_raster.GetRasterBand(2).SetColorInterpretation(gdal.GCI_GreenBand)
        output_raster.GetRasterBand(2).FlushCache()
        if nodata:
            output_raster.GetRasterBand(2).SetNoDataValue(nodata)

        output_raster.GetRasterBand(3).WriteArray(pixels[:,:,2].astype('uint8'))
        output_raster.GetRasterBand(3).SetColorInterpretation(gdal.GCI_BlueBand)
        output_raster.GetRasterBand(3).FlushCache()
        if nodata:
            output_raster.GetRasterBand(3).SetNoDataValue(nodata)

    elif channels > 1:
        # TODO: Something wonky w/ uint8s --> ending up w/ lots of gaps in data (white pixels)
        for chan in range(channels):
            band = chan + 1
            output_raster.GetRasterBand(band).WriteArray(pixels[:,:,chan].astype('uint8'))
            output_raster.GetRasterBand(band).FlushCache()
            if nodata:
                output_raster.GetRasterBand(band).SetNoDataValue(nodata)
    else:
        # single channel image, e.g. temperature
        output_raster.GetRasterBand(1).WriteArray(pixels)
        output_raster.GetRasterBand(1).FlushCache()
        if nodata:
            output_raster.GetRasterBand(1).SetNoDataValue(nodata)

    output_raster = None


def prepare_metadata_for_geotiff(extractor_info=None, terra_md=None):
    """Create geotiff-embedded metadata from extractor_info and terraref metadata pieces.

        Keyword arguments:
        extractor_info -- details about extractor if applicable
        system_md -- cleaned TERRA-REF metadata
    """
    extra_metadata = {}

    if (terra_md != None):
        extra_metadata["datetime"] = str(terra_md["gantry_variable_metadata"]["datetime"])
        extra_metadata["sensor_id"] = str(terra_md["sensor_fixed_metadata"]["sensor_id"])
        extra_metadata["sensor_url"] = str(terra_md["sensor_fixed_metadata"]["url"])
        experiment_names = []
        for e in terra_md["experiment_metadata"]:
            experiment_names.append(e["name"])
        terra_md["experiment_name"] = ", ".join(experiment_names)

    if (extractor_info != None):
        extra_metadata["extractor_name"] = str(extractor_info.get("name", ""))
        extra_metadata["extractor_version"] = str(extractor_info.get("version", ""))
        extra_metadata["extractor_author"] = str(extractor_info.get("author", ""))
        extra_metadata["extractor_description"] = str(extractor_info.get("description", ""))
        if "repository" in extractor_info and "repUrl" in extractor_info["repository"]:
            extra_metadata["extractor_repo"] = str(extractor_info["repository"]["repUrl"])
        else:
            extra_metadata["extractor_repo"] = ""

    return extra_metadata


def create_netcdf(pixels, out_path, scaled=False):
    """Generate output netCDF file given an input numpy pixel array.

            Keyword arguments:
            pixels -- 2-dimensional numpy array of pixel values
            out_path -- path to GeoTIFF to be created
            scaled -- whether to scale PNG output values based on pixels min/max
        """
    dimensions = numpy.shape(pixels)
    if len(dimensions) == 2:
        nrows, ncols = dimensions
        channels = 1
    else:
        nrows, ncols, channels = dimensions

    out_nc = Dataset(out_path, "w", format="NETCDF4")
    out_nc.createDimension('band',  channels) # only 1 band for mask
    out_nc.createDimension('x', ncols)
    out_nc.createDimension('y', nrows)

    mask = out_nc.createVariable('soil_mask','f8',('band', 'x', 'y'))
    mask[:] = pixels

    out_nc.close()


def create_image(pixels, out_path, scaled=False):
    """Generate output JPG/PNG file given an input numpy pixel array.

            Keyword arguments:
            pixels -- 2-dimensional numpy array of pixel values
            out_path -- path to GeoTIFF to be created
            scaled -- whether to scale PNG output values based on pixels min/max
        """
    if scaled:
        # e.g. flirIrCamera
        Gmin = pixels[pixels>0].min()
        Gmax = pixels.max()
        scaled_px = (pixels-Gmin)/(Gmax - Gmin)
        plt.imsave(out_path, cm.get_cmap('jet')(scaled_px))
    else:
        # e.g. PSII
        # TODO: Can we make these use same library?
        # TODO: plt.imsave(out_path, pixels)
        Image.fromarray(pixels).save(out_path)



"""Spatial

This module provides useful methods for spatial referencing.
"""

import os
import json
import logging
import subprocess
import numpy as np
import yaml
import laspy
import osr
from osgeo import gdal, ogr
import utm

def convert_geometry(geometry, new_spatialreference):
    """Converts the geometry to the new spatial reference if possible

    geometry - The geometry to transform
    new_spatialreference - The spatial reference to change to

    Returns:
        The transformed geometry or the original geometry. If either the
        new Spatial Reference parameter is None, or the geometry doesn't
        have a spatial reference, then the original geometry is returned.
    """
    if not new_spatialreference or not geometry:
        return geometry

    return_geometry = geometry
    try:
        geom_sr = geometry.GetSpatialReference()
        if geom_sr and not new_spatialreference.IsSame(geom_sr):
            transform = osr.CreateCoordinateTransformation(geom_sr, new_spatialreference)
            new_geom = geometry.Clone()
            if new_geom:
                new_geom.Transform(transform)
                return_geometry = new_geom
    except Exception as ex:
        logging.warning("Exception caught while transforming geometries: " + str(ex))
        logging.warning("    Returning original geometry")

    return return_geometry

def convert_json_geometry(geojson, new_spatialreference):
    """Converts geojson geometry to new spatial reference system and
       returns the new geometry as geojson.

    geojson(str): The geometry to transform
    new_spatialreference - The spatial reference to change to

    Returns:
        The transformed geometry as geojson or the original geojson. If
        either the new Spatial Reference parameter is None, or the geojson
        doesn't have a spatial reference, or the geojson isn't a
        valid geometry, then the original geojson is returned.
    """
    if not new_spatialreference or not geojson:
        return geojson

    geom_yaml = yaml.safe_load(geojson)
    geometry = ogr.CreateGeometryFromJson(json.dumps(geom_yaml))

    if not geometry:
        return geojson

    new_geometry = convert_geometry(geometry, new_spatialreference)
    if not new_geometry:
        return geojson

    try:
        return geometry_to_geojson(new_geometry)
    except Exception as ex:
        logging.warning("Exception caught while transforming geojson: " + str(ex))
        logging.warning("    Returning original geojson")

    return geojson

def calculate_bounding_box(gps_bounds, z_value=0):
    """Given a set of GPS boundaries, return array of 4 vertices representing the polygon.

    gps_bounds -- (lat(y) min, lat(y) max, long(x) min, long(x) max)

    Returns:
        List of 4 polygon vertices [ul, ur, lr, ll] without z values.
    """

    return [
        (gps_bounds[2], gps_bounds[1], z_value), # upper-left
        (gps_bounds[3], gps_bounds[1], z_value), # upper-right
        (gps_bounds[3], gps_bounds[0], z_value), # lower-right
        (gps_bounds[2], gps_bounds[0], z_value)  # lower-left
    ]


def calculate_centroid(gps_bounds):
    """Given a set of GPS boundaries, return lat/lon of centroid.

    gps_bounds -- (lat(y) min, lat(y) max, long(x) min, long(x) max)

    Returns:
        Tuple of (lat, lon) representing centroid
    """

    return (
        gps_bounds[0] + float(gps_bounds[1] - gps_bounds[0])/2,
        gps_bounds[2] + float(gps_bounds[3] - gps_bounds[2])/2,
    )


def calculate_centroid_from_wkt(wkt):
    """Given WKT, return lat/lon of centroid.

    wkt -- string

    returns:
        Tuple of (lat, lon) representing centroid
    """

    loc_geom = ogr.CreateGeometryFromWkt(wkt)
    return (
        loc_geom.Centroid().GetX(),
        loc_geom.Centroid().GetY()
    )


def calculate_gps_bounds(metadata, sensor="stereoTop"):
    """Extract bounding box geometry, depending on sensor type.

        Gets geometry from metadata for center position and FOV, applies some
        sensor-specific transformations to those values, then uses them in formula
        to determine bounding box of image.

        Returns:
            tuple of GeoTIFF coordinates, each one as:
            (lat(y) min, lat(y) max, long(x) min, long(x) max)
    """
    gantry_x, gantry_y, gantry_z, cambox_x, cambox_y, cambox_z, fov_x, fov_y = geom_from_metadata(metadata)

    center_position = ( float(gantry_x) + float(cambox_x),
                        float(gantry_y) + float(cambox_y),
                        float(gantry_z) + float(cambox_z) )
    cam_height = center_position[2]

    if sensor=="stereoTop":
        var_se = float(metadata['sensor_fixed_metadata']['slope_estimation'])
        var_rho = float(metadata['sensor_fixed_metadata']['rail_height_offset'])
        var_sofc = float(metadata['sensor_fixed_metadata']['stereo_offsets_from_center'])

        # Use height of camera * slope_estimation to estimate expected canopy height
        predicted_plant_height = var_se * cam_height
        # Subtract expected plant height from (cam height + rail height offset) to get canopy height
        cam_height_above_canopy = float(cam_height + var_rho - predicted_plant_height)
        fov_x = float(fov_x) * (cam_height_above_canopy/2)
        fov_y = float(fov_y) * (cam_height_above_canopy/2)
        # Account for experimentally determined distance from center to each stereo lens for left/right
        left_position = [center_position[0]-var_sofc, center_position[1], center_position[2]]
        right_position = [center_position[0]+var_sofc, center_position[1], center_position[2]]
        # Return two separate bounding boxes for left/right
        left_gps_bounds = _get_bounding_box_with_formula(left_position, [fov_x, fov_y])
        right_gps_bounds = _get_bounding_box_with_formula(right_position, [fov_x, fov_y])
        return { "left" : left_gps_bounds, "right" : right_gps_bounds }

    elif sensor=="flirIrCamera":
        cam_height_above_canopy = cam_height + float(metadata['sensor_fixed_metadata']['rail_height_offset'])
        fov_x = float(fov_x) * (cam_height_above_canopy/2)
        fov_y = float(fov_y) * (cam_height_above_canopy/2)

    elif sensor=='scanner3DTop':
        # Default geom refers to west side, so get east side cambox as well
        gx, gy, gz, e_cambox_x, e_cambox_y, e_cambox_z, fx, fy = geom_from_metadata(metadata, 'east')

        # Swap X and Y because we rotate 90 degress
        fov_x = float(fov_y) if fov_y else 0
        scan_distance = float(metadata['sensor_variable_metadata']['scan_distance_mm'])/1000
        fov_y = scan_distance
        scandirection = int(metadata['sensor_variable_metadata']['scan_direction'])

        # TODO: These constants should live in fixed metadata once finalized
        if scandirection == 0: # Negative scan
            west_position = ( float(gantry_x) + float(cambox_x) + 0.082,
                              float(gantry_y) + float(2*float(cambox_y)) - scan_distance/2 - 4.363, #Might be less than this
                              float(gantry_z) + float(cambox_z) )

            east_position = ( float(gantry_x) + float(e_cambox_x) + 0.082,
                              float(gantry_y) + float(2*float(e_cambox_y)) - scan_distance/2 - 0.354,
                              float(gantry_z) + float(e_cambox_z) )
        else: # Positive scan
            west_position = ( float(gantry_x) + float(cambox_x) + 0.082,
                              float(gantry_y) + float(2*float(cambox_y)) + scan_distance/2 - 4.23,
                              float(gantry_z) + float(cambox_z) )

            east_position = ( float(gantry_x) + float(e_cambox_x) + 0.082,
                              float(gantry_y) + float(2*float(e_cambox_y)) + scan_distance/2 + 0.4,
                              float(gantry_z) + float(e_cambox_z) )

        east_gps_bounds = _get_bounding_box_with_formula(east_position, [fov_x, fov_y])
        west_gps_bounds = _get_bounding_box_with_formula(west_position, [fov_x, fov_y])
        return { "east" : east_gps_bounds, "west" : west_gps_bounds }

    else:
        fov_x = float(fov_x) if fov_x else 0
        fov_y = float(fov_y) if fov_y else 0

    return { sensor : _get_bounding_box_with_formula(center_position, [fov_x, fov_y]) }


def centroid_from_geojson(geojson):
    """Return centroid lat/lon of a geojson object."""
    geom_poly = ogr.CreateGeometryFromJson(geojson)
    centroid = geom_poly.Centroid()

    return centroid.ExportToJson()


def clip_las(las_path, tuples, out_path, merged_path=None):
    """Clip LAS file to polygon.

    Args:
      las_path (str): path to pointcloud file
      geojson (str): geoJSON bounds received from get_site_boundaries() in betydb.py
      out_path: output file to write
    """
    utm = tuples_to_utm(tuples)
    bounds_str = "([%s, %s], [%s, %s])" % (utm[2], utm[3], utm[0], utm[1])

    pdal_dtm = out_path.replace(".las", "_dtm.json")
    with open(pdal_dtm, 'w') as dtm:
        dtm.write("""{
            "pipeline": [
                "%s",
                {
                    "type": "filters.crop",
                    "bounds": "%s"
                },
                {
                    "type": "writers.las",
                    "filename": "%s"
                }
            ]
        }""" % (las_path, bounds_str, out_path))

    cmd = 'pdal pipeline "%s"' % pdal_dtm
    subprocess.call([cmd], shell=True)
    os.remove(pdal_dtm)

    if merged_path:
        if os.path.isfile(merged_path):
            cmd = 'pdal merge "%s" "%s" "%s"' % (out_path, merged_path, merged_path)
            subprocess.call([cmd], shell=True)
        else:
            os.rename(out_path, merged_path)


def clip_raster(rast_path, bounds, out_path=None, nodata=-9999, compress=False):
    """Clip raster to polygon.

    Args:
      rast_path (str): path to raster file
      bounds (tuple): (min_y, max_y, min_x, max_x)
      out_path: if provided, where to save as output file
      nodata: the no data value

    Returns: (numpy array, GeoTransform)

    Notes: Oddly, the "features path" can be either a filename
      OR a geojson string. GDAL seems to figure it out and do
      the right thing.

      From http://karthur.org/2015/clipping-rasters-in-python.html
    """

    if not out_path:
        out_path = "temp.tif"

    # Clip raster to GDAL and read it to numpy array
    coords = "%s %s %s %s" % (bounds[2], bounds[1], bounds[3], bounds[0])
    if compress:
        cmd = 'gdal_translate -projwin %s "%s" "%s"' % (coords, rast_path, out_path)
    else:
        cmd = 'gdal_translate -co COMPRESS=LZW -projwin %s "%s" "%s"' % (coords, rast_path, out_path)
    subprocess.call(cmd, shell=True, stdout=open(os.devnull, 'wb'))
    out_px = np.array(gdal.Open(out_path).ReadAsArray())

    if np.count_nonzero(out_px) > 0:
        if out_path == "temp.tif":
            os.remove(out_path)
        return out_px
    else:
        os.remove(out_path)
        return None


def find_plots_intersect_boundingbox(bounding_box, all_plots, fullmac=True):
    """Take a list of plots from BETY and return only those overlapping bounding box.

    fullmac -- only include full plots (omit KSU, omit E W partial plots)

    """
    bbox_poly = ogr.CreateGeometryFromJson(str(bounding_box))
    bb_sr = bbox_poly.GetSpatialReference()
    intersecting_plots = dict()

    for plotname in all_plots:
        if fullmac and (plotname.find("KSU") > -1 or plotname.endswith(" E") or plotname.endswith(" W")):
            continue

        bounds = all_plots[plotname]

        yaml_bounds = yaml.safe_load(bounds)
        current_poly = ogr.CreateGeometryFromJson(json.dumps(yaml_bounds))

        # Check for a need to convert coordinate systems
        check_poly = current_poly
        if bb_sr:
            poly_sr = current_poly.GetSpatialReference()
            if poly_sr and not bb_sr.IsSame(poly_sr):
                # We need to convert to the same coordinate system before an intersection
                check_poly = convert_geometry(current_poly, bb_sr)
                transform = osr.CreateCoordinateTransformation(poly_sr, bb_sr)
                new_poly = current_poly.Clone()
                if new_poly:
                    new_poly.Transform(transform)
                    check_poly = new_poly

        intersection_with_bounding_box = bbox_poly.Intersection(check_poly)

        if intersection_with_bounding_box is not None:
            intersection = json.loads(intersection_with_bounding_box.ExportToJson())
            if 'coordinates' in intersection and len(intersection['coordinates']) > 0:
                intersecting_plots[plotname] = bounds

    return intersecting_plots


def geojson_to_tuples(bounding_box):
    """
    Given a GeoJSON polygon, returns in tuple format
     (lat(y) min, lat(y) max, long(x) min, long(x) max)
    """
    lat_max = bounding_box["coordinates"][0][0]
    long_min = bounding_box["coordinates"][0][1]
    long_max = bounding_box["coordinates"][1][1]
    lat_min = bounding_box["coordinates"][2][0]
    
    return (lat_min, lat_max, long_min, long_max)


def geojson_to_tuples_betydb(bounding_box):
    """Convert GeoJSON from BETYdb to
        ( lat (y) min, lat (y) max,
          long (x) min, long (x) max) for geotiff creation"""
    min_x, min_y, max_x, max_y  = None, None, None, None

    if isinstance(bounding_box, dict):
        bounding_box = bounding_box["coordinates"]

    for coord in bounding_box[0][0]:
        if not min_x or coord[0] < min_x:
            min_x = coord[0]
        if not max_x or coord[0] > max_x:
            max_x = coord[0]
        if not min_y or coord[1] < min_y:
            min_y = coord[1]
        if not max_y or coord[1] > max_y:
            max_y = coord[1]

    return (min_y, max_y, min_x, max_x)


def geom_from_metadata(metadata, side='west'):
    """Parse location elements from metadata.

        Returns:
        tuple of location information: (
            location of scannerbox x, y, z
            location offset of sensor in scannerbox in x, y, z
            field-of-view of camera in x, y dimensions
        )
    """
    gantry_x, gantry_y, gantry_z = None, None, None
    cambox_x, cambox_y, cambox_z = None, None, None
    fov_x, fov_y = None, None

    if 'terraref_cleaned_metadata' in metadata and metadata['terraref_cleaned_metadata']:
        gv_meta = metadata['gantry_variable_metadata']
        gantry_x = gv_meta['position_m']['x'] if 'x' in gv_meta['position_m'] else gantry_x
        gantry_y = gv_meta['position_m']['y'] if 'y' in gv_meta['position_m'] else gantry_y
        gantry_z = gv_meta['position_m']['z'] if 'z' in gv_meta['position_m'] else gantry_z

        if 'sensor_fixed_metadata' in metadata:
            sf_meta = metadata['sensor_fixed_metadata']

        # LOCATION IN CAMERA BOX
        if 'location_in_camera_box_m' in sf_meta:
            cambox_x = sf_meta['location_in_camera_box_m']['x']
            cambox_y = sf_meta['location_in_camera_box_m']['y']
            cambox_z = sf_meta['location_in_camera_box_m']['z'] if 'z' in sf_meta['location_in_camera_box_m'] else 0
        elif side=='west' and 'scanner_west_location_in_camera_box_m' in sf_meta:
            cambox_x = sf_meta['scanner_west_location_in_camera_box_m']['x']
            cambox_y = sf_meta['scanner_west_location_in_camera_box_m']['y']
            cambox_z = sf_meta['scanner_west_location_in_camera_box_m']['z']
        elif side=='east' and 'scanner_east_location_in_camera_box_m' in sf_meta:
            cambox_x = sf_meta['scanner_east_location_in_camera_box_m']['x']
            cambox_y = sf_meta['scanner_east_location_in_camera_box_m']['y']
            cambox_z = sf_meta['scanner_east_location_in_camera_box_m']['z']

        # FIELD OF VIEW (FOV)
        for fov_field in ['field_of_view_m', 'field_of_view_at_2m_m', 'field_of_view_degrees']:
            if fov_field in sf_meta:
                fov_x = sf_meta[fov_field]['x'] if 'x' in sf_meta[fov_field] else fov_x
                fov_y = sf_meta[fov_field]['y'] if 'y' in sf_meta[fov_field] else fov_y

    return (gantry_x, gantry_y, gantry_z, cambox_x, cambox_y, cambox_z, fov_x, fov_y)


def get_las_extents(fname):
    """Calculate the extent of the given pointcloud and return as GeoJSON."""
    lasinfo = laspy.base.Reader(fname, 'r')
    min = lasinfo.get_header().min
    max = lasinfo.get_header().max

    min_latlon = utm.to_latlon(min[1], min[0], 12, 'S')
    max_latlon = utm.to_latlon(max[1], max[0], 12, 'S')

    return {
        "type": "Polygon",
        "coordinates": [[
            [min_latlon[1], min_latlon[0]],
            [min_latlon[1], max_latlon[0]],
            [max_latlon[1], max_latlon[0]],
            [max_latlon[1], min_latlon[0]],
            [min_latlon[1], min_latlon[0]]
        ]]
    }


def get_raster_extents(fname):
    """Calculate the extent and the center of the given raster."""
    src = gdal.Open(fname)
    ulx, xres, xskew, uly, yskew, yres = src.GetGeoTransform()
    lrx = ulx + (src.RasterXSize * xres)
    lry = uly + (src.RasterYSize * yres)
    extent = (ulx, lry, lrx, uly)
    center = ((ulx+lrx)/2, (uly+lry)/2)

    return (extent, center)


def utm_to_latlon(utm_x, utm_y):
    """Convert coordinates from UTM 12N to lat/lon"""

    # Get UTM information from southeast corner of field
    SE_utm = utm.from_latlon(33.07451869, -111.97477775)
    utm_zone = SE_utm[2]
    utm_num  = SE_utm[3]

    return utm.to_latlon(utm_x, utm_y, utm_zone, utm_num)


def scanalyzer_to_latlon(gantry_x, gantry_y):
    """Convert coordinates from gantry to lat/lon"""
    utm_x, utm_y = scanalyzer_to_utm(gantry_x, gantry_y)
    return utm_to_latlon(utm_x, utm_y)


def scanalyzer_to_utm(gantry_x, gantry_y):
    """Convert coordinates from gantry to UTM 12N"""

    # TODO: Hard-coded
    # Linear transformation coefficients
    ay = 3659974.971; by = 1.0002; cy = 0.0078;
    ax = 409012.2032; bx = 0.009; cx = - 0.9986;

    utm_x = ax + (bx * gantry_x) + (cx * gantry_y)
    utm_y = ay + (by * gantry_x) + (cy * gantry_y)

    return utm_x, utm_y


def tuples_to_geojson(bounds):
    """
    Given bounding box in tuple format, returns GeoJSON polygon

    Input bounds: (lat(y) min, lat(y) max, long(x) min, long(x) max)
    """
    lat_min = bounds[0]
    lat_max = bounds[1]
    long_min = bounds[2]
    long_max = bounds[3]

    bounding_box = {}
    bounding_box["type"] = "Polygon"
    bounding_box["coordinates"]  =  [
        [lat_max, long_min], # NW
        [lat_max, long_max], # NE
        [lat_min, long_max], # SE
        [lat_min, long_min]  # SW
    ]

    return bounding_box


def tuples_to_utm(bounds):
    """Given bounding box in tuple format, returns UTM equivalent

    Input bounds: (lat(y) min, lat(y) max, long(x) min, long(x) max)
    """

    min = utm.from_latlon(bounds[0], bounds[2])
    max = utm.from_latlon(bounds[1], bounds[3])

    return (min[0], max[0], min[1], max[1])


def wkt_to_geojson(wkt):
    geom = ogr.CreateGeometryFromWkt(wkt)
    return geom.ExportToJson()

def geometry_to_geojson(geom, alt_coord_type=None, alt_coord_code=None):
    """Converts a geometry to geojson.
    Args:
        geom(ogr geometry): The geometry to convert to JSON
        alt_coord_type(str): the alternate geographic coordinate system type if geometry doesn't have one defined
        alt_coord_code(str): the alternate geographic coordinate system associated with the type
    Returns:
        The geojson string for the geometry
    Note:
        If the geometry doesn't have a spatial reference associated with it, both the default
        coordinate system type and code must be specified for a coordinate system to be assigned to
        the returning JSON. The original geometry is left unaltered.
    """
    ref_sys = geom.GetSpatialReference()
    geom_json = json.loads(geom.ExportToJson())
    if not ref_sys:
        if alt_coord_type and alt_coord_code:
            # Coming from BETYdb without a coordinate system we assume EPSG:4326
            geom_json['crs'] = {'type': str(alt_coord_type), 'properties': {'code': str(alt_coord_code)}}
    else:
        geom_json['crs'] = {
            'type': ref_sys.GetAttrValue("AUTHORITY", 0),
            'properties': {
                'code': ref_sys.GetAttrValue("AUTHORITY", 1)
            }
        }

    return json.dumps(geom_json)


# PRIVATE -------------------------------------
def _get_bounding_box_with_formula(center_position, fov):
    """Convert scannerbox center position & sensor field-of-view to actual bounding box

        Linear transformation formula adapted from:
        https://terraref.gitbooks.io/terraref-documentation/content/user/geospatial-information.html

        Returns:
            tuple of coordinates as: (  lat (y) min, lat (y) max,
                                        long (x) min, long (x) max )
    """

    # min/max bounding box x,y values
    y_w = center_position[1] + fov[1]/2
    y_e = center_position[1] - fov[1]/2
    x_n = center_position[0] + fov[0]/2
    x_s = center_position[0] - fov[0]/2
    # coordinates of northwest bounding box vertex
    bbox_nw_latlon = scanalyzer_to_latlon(x_n, y_w)
    # coordinates if southeast bounding box vertex
    bbox_se_latlon = scanalyzer_to_latlon(x_s, y_e)

    # TODO: Hard-coded
    lon_shift = 0.000020308287
    lat_shift = 0.000015258894
    return ( bbox_se_latlon[0] - lat_shift,
             bbox_nw_latlon[0] - lat_shift,
             bbox_nw_latlon[1] + lon_shift,
             bbox_se_latlon[1] + lon_shift )



"""Clowder API

This module provides simple wrappers around the clowder Geostreams API
"""

import json
import logging
import requests

from terrautils.betydb import get_sites_by_latlon
from terrautils.spatial import wkt_to_geojson


def create_sensor(connector, host, key, sensorname, geom, type, region):
    """Create a new sensor in Geostreams.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    sensorname -- name of new sensor to create
    geom -- GeoJSON object of sensor geometry
    type -- JSON object with {"id", "title", and "sensorType"}
    region -- region of sensor
    """

    logger = logging.getLogger(__name__)

    body = {
        "name": sensorname,
        "type": "Point",
        "geometry": geom,
        "properties": {
            "popupContent": sensorname,
            "type": type,
            "name": sensorname,
            "region": region
        }
    }

    url = "%sapi/geostreams/sensors?key=%s" % (host, key)

    result = requests.post(url, headers={'Content-type': 'application/json'},
                           data=json.dumps(body),
                           verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    sensorid = result.json()['id']
    logger.debug("sensor id = [%s]", sensorid)

    return sensorid


def create_stream(connector, host, key, streamname, sensorid, geom, properties={}):
    """Create a new stream in Geostreams.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    streamname -- name of new stream to create
    sensorid -- id of sensor to attach stream to
    geom -- GeoJSON object of sensor geometry
    properties -- JSON object with any desired properties
    """

    logger = logging.getLogger(__name__)

    body = {
        "name": streamname,
        "type": "Feature",
        "geometry": geom,
        "properties": properties,
        "sensor_id": str(sensorid)
    }

    url = "%sapi/geostreams/streams?key=%s" % (host, key)

    result = requests.post(url, headers={'Content-type': 'application/json'},
                           data=json.dumps(body),
                           verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    streamid = result.json()['id']
    logger.debug("stream id = [%s]", streamid)

    return streamid


def create_datapoint(connector, host, key, streamid, geom, starttime, endtime, properties={}):
    """Create a new datapoint in Geostreams.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    streamid -- id of stream to attach datapoint to
    geom -- GeoJSON object of sensor geometry
    starttime -- start time, in format 2017-01-25T09:33:02-06:00
    endtime -- end time, in format 2017-01-25T09:33:02-06:00
    properties -- JSON object with any desired properties
    """

    logger = logging.getLogger(__name__)

    body = {
        "start_time": starttime,
        "end_time": endtime,
        "type": "Point",
        "geometry": geom,
        "properties": properties,
        "stream_id": str(streamid)
    }

    url = '%sapi/geostreams/datapoints?key=%s' % (host, key)

    result = requests.post(url, headers={'Content-type': 'application/json'},
                           data=json.dumps(body),
                           verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    dpid = result.json()['id']
    logger.debug("datapoint id = [%s]", dpid)

    return dpid


def create_datapoints(connector, host, key, streamid, datapoint_list):
    """Create a new batch of datapoints in Geostreams.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    streamid -- id of stream to attach datapoint to
    datapoint_list -- list of datapoint objects with start & end time, type, geometry, properties
    """

    body = {
        "datapoints": datapoint_list,
        "stream_id": str(streamid)
    }

    url = '%sapi/geostreams/datapoints/bulk?key=%s' % (host, key)

    result = requests.post(url, headers={'Content-type': 'application/json'},
                           data=json.dumps(body),
                           verify=connector.ssl_verify if connector else True)
    result.raise_for_status()


def create_datapoint_with_dependencies(connector, host, key, streamprefix, latlon, starttime, endtime,
                                       metadata={}, filter_date='', geom=None, plot_name=None):
    """Create a new datapoint in Geostreams. Will create sensor and stream as necessary.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    streamid -- id of stream to attach datapoint to
    latlon -- [latitude, longitude] tuple of datapoint location
    starttime -- start time, in format 2017-01-25T09:33:02-06:00
    endtime -- end time, in format 2017-01-25T09:33:02-06:00
    metadata -- JSON object with any desired properties
    filter_date -- date used to restrict number of sites returned from BETYdb
    geom -- geometry for datapoint (use plot if not provided)
    plot_name -- name of plot to map datapoint into if possible, otherwise query BETY
    """

    # SENSOR is the plot
    matched_sites = {}
    if plot_name:
        # If provided a plot name, see if the sensor exists before going into more expensive logic
        sensor_data = get_sensor_by_name(connector, host, key, plot_name)
        if sensor_data:
            matched_sites[sensor_data['id']] = {
                "name": plot_name,
                "geom": sensor_data['geometry']
            }

    if matched_sites == {}:
        # If we don't have existing sensor to use quickly, we must query geographically
        sitelist = get_sites_by_latlon(latlon, filter_date)
        for s in sitelist:
            plot_name = s['sitename']
            plot_geom = json.loads(wkt_to_geojson(s['geometry']))

            # Get existing sensor with this plot name from geostreams, or create if it doesn't exist
            sensor_data = get_sensor_by_name(connector, host, key, plot_name)
            if not sensor_data:
                sensor_id = create_sensor(connector, host, key, plot_name, plot_geom,
                                          {"id": "MAC Field Scanner", "title": "MAC Field Scanner", "sensorType": 4},
                                          "Maricopa")
                matched_sites[sensor_id] = {"name": plot_name, "geom": plot_geom}
            else:
                sensor_id = sensor_data['id']
                matched_sites[sensor_id] = {"name": plot_name, "geom": plot_geom}

    for sensor_id in matched_sites:
        plot_geom = matched_sites[sensor_id]["geom"]
        stream_name = "%s (%s)" % (streamprefix, sensor_id)
        stream_data = get_stream_by_name(connector, host, key, stream_name)
        if not stream_data:
            stream_id = create_stream(connector, host, key, stream_name, sensor_id, plot_geom)
        else:
            stream_id = stream_data['id']

        logging.info("posting datapoint to stream %s" % stream_id)
        if not geom:
            geom = plot_geom
        create_datapoint(connector, host, key, stream_id, geom,
                         starttime, endtime, metadata)


def get_sensor_by_name(connector, host, key, sensorname):
    """Get sensor by name from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    sensorname -- name of sensor to search for
    """

    logger = logging.getLogger(__name__)

    url = "%sapi/geostreams/sensors?sensor_name=%s&key=%s" % (host, sensorname, key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    for sens in result.json():
        if 'name' in sens and sens['name'] == sensorname:
            logger.debug("found sensor '%s' = [%s]" % (sensorname, sens['id']))
            return sens

    return None


def get_sensors_by_circle(connector, host, key, lon, lat, radius=0):
    """Get sensor by coordinate from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    lon -- longitude of point
    lat -- latitude of point
    radius -- distance in meters around point to search
    """

    logger = logging.getLogger(__name__)

    url = "%sapi/geostreams/sensors?geocode=%s,%s,%s&key=%s" % (host, lat, lon, radius, key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    # Return first sensor
    jbody = result.json()
    if len(jbody) > 0:
        return jbody
    else:
        return None


def get_sensors_by_polygon(connector, host, key, coord_list):
    """Get sensor by coordinate from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    coord_list -- list of (lon/lat) coordinate pairs forming polygon vertices
    """

    logger = logging.getLogger(__name__)

    coord_strings = [str(i) for i in coord_list]
    url = "%sapi/geostreams/sensors?geocode=%s&key=%s" % (host, ','.join(coord_strings), key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    # Return first sensor
    jbody = result.json()
    if len(jbody) > 0:
        return jbody
    else:
        return None


def get_stream_by_name(connector, host, key, streamname):
    """Get stream by name from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    streamname -- name of stream to search for
    """

    logger = logging.getLogger(__name__)

    url = "%sapi/geostreams/streams?stream_name=%s&key=%s" % (host, streamname, key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    for strm in result.json():
        if 'name' in strm and strm['name'] == streamname:
            logger.debug("found stream '%s' = [%s]" % (streamname, strm['id']))
            return strm

    return None


def get_streams_by_circle(connector, host, key, lon, lat, radius=0):
    """Get stream by coordinate from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    lon -- longitude of point
    lat -- latitude of point
    radius -- distance in meters around point to search
    """

    logger = logging.getLogger(__name__)

    url = "%sapi/geostreams/stream?geocode=%s,%s,%s&key=%s" % (host, lat, lon, radius, key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    jbody = result.json()
    if len(jbody) > 0:
        return jbody
    else:
        return None


def get_streams_by_polygon(connector, host, key, coord_list):
    """Get stream by coordinate from Geostreams, or return None.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    coord_list -- list of (lon/lat) coordinate pairs forming polygon vertices
    """

    logger = logging.getLogger(__name__)

    coord_strings = [str(i) for i in coord_list]
    url = "%sapi/geostreams/stream?geocode=%s&key=%s" % (host, ','.join(coord_strings), key)

    result = requests.get(url,
                          verify=connector.ssl_verify if connector else True)
    result.raise_for_status()

    jbody = result.json()
    if len(jbody) > 0:
        return jbody
    else:
        return None



"""Extractors

This module provides useful reference methods for extractors.
"""

import datetime
import time
import logging
import json
import copy
import os
import re
import sys
import requests
import yaml
from urllib3.filepost import encode_multipart_formdata

from pyclowder.extractors import Extractor
from pyclowder.datasets import get_file_list, download_metadata as download_dataset_metadata
from terrautils.influx import Influx, add_arguments as add_influx_arguments
from terrautils.metadata import get_terraref_metadata, pipeline_get_metadata, \
                get_season_and_experiment
from terrautils.sensors import Sensors, STATIONS, add_arguments as add_sensor_arguments
from terrautils.users import get_dataset_username, find_user_name


logging.basicConfig(format='%(asctime)s %(message)s')

DEFAULT_EXPERIMENT_JSON_FILENAME = 'experiment.yaml'

class __internal__(object):
    """Class for functions intended for internal use only for this file
    """
    def __init__(self):
        """Performs initialization of class instance
        """
        pass

    @staticmethod
    def case_insensitive_find(data, key, default_value=None):
        """Performs a shallow case insensitive lookup of the key in the data.
        Args:
            data(dict): The haystack to search
            key(str): The key to find
            default_value: optional parameter returned if the key isn't found
        Return:
            The found value is returned, None if there's an error, and default_value if it
            isn't found
        """
        if not data or not key:
            return None

        # Return exact match value [assuming this is faster than iterating]
        if key in data:
            return data[key]

        # Find the key
        lower_key = key.lower()
        for item in data:
            if item.lower() == lower_key:
                return data[item]

        return default_value

    @staticmethod
    def merge_experiment_json(start_json, more_json):
        """Performs a merge on the json dictionary data without overwriting
           complex upper level keys. Lists and tuples are replaced if keys match.
           Dictionaries are merged with only lowest level keys being replaced when
           there's a key match.
        Args:
            start_json(dict): the starting json to work with
            more_json(dict or list): another json dict or a list of dicts
        Return:
            A unified dictionary
        Notes:
            All the parameter JSON objects are left intact. New JSON objects are
            created as needed.
        """
        return_json = dict(start_json)
        if more_json is None:
            return return_json

        def _recurse_item(item_old, item_new):
            """Internal function to merge_experiment_json() function"""
            # Use item types to decide what to do
            if not isinstance(item_old, type(item_new)):
                return item_new
            if not isinstance(item_old, dict):
                return item_new

            # Setup to merge
            new_json = dict(item_old)
            cur_keys = new_json.keys()
            new_keys = item_new.keys()

            # Merge or replace values for same-name keys
            for one_key in cur_keys:
                if one_key in new_keys:
                    new_json[one_key] = _recurse_item(new_json[one_key], item_new[one_key])

            # Add new items from the new JSON into what we're building up
            for one_key in new_keys:
                if not one_key in cur_keys:
                    new_json[one_key] = item_new[one_key]

            return new_json

        # Loop through our json objects
        if not isinstance(more_json, list):
            more_json = [more_json]

        for one_json in more_json:
            # We only deal with json dictionary objects
            if isinstance(one_json, dict):
                return_json = _recurse_item(return_json, one_json)

        return return_json

    @staticmethod
    def is_string(to_check):
        """Performs checks to ensure the parameter is a string type and also not None
        Keyword Arguments:
            to_check: The value to check for string-ness
        Return:
            Returns True if the parameter is a string and False if None or not a string
        """
        if not to_check or not isinstance(to_check, str):
            return False
        if sys.version_info[0] < 3:
            if not isinstance(to_check, unicode):
                return False
        return True

    @staticmethod
    def prep_string_for_filename(source):
        """Perpares a string to be used as part of a file path or file name
        Keyword Arguments:
            source(str): the source string
        Return:
            Returns the formatting string. The original value is returned if
            the parameter can't be formatted
        """
        try:
            formatted = source.replace('/', '.').replace('\\', '.').replace('&', '.').replace('*', '.') \
                              .replace("'", '.').replace('"', '.').replace('`', '.') \
                              .replace(' ', '_').replace('\t', '_').replace('\r', '_')
        except Exception as ex:
            logging.warning("Exception caught while preparing filename: " + str(ex))
            logging.warning("    returning original parameter")
            formatted = source

        return formatted

def add_arguments(parser):

    # TODO: Move defaults into a level-based dict
    parser.add_argument('--clowderspace',
                        default=os.getenv('CLOWDER_SPACE', "58da6b924f0c430e2baa823f"),
                        help='sets the default Clowder space for creating new things')

    parser.add_argument('--overwrite',
                        default=False,
                        action='store_true',
                        help='enable overwriting of existing files')

    parser.add_argument('--debug', '-d', action='store_const',
                        default=logging.INFO, const=logging.DEBUG,
                        help='enable debugging (default=WARN)')

    parser.add_argument('--clowder_user',
                        default=os.getenv('CLOWDER_USER', "terrarefglobus+uamac@ncsa.illinois.edu"),
                        help='clowder user to use when creating new datasets')

    parser.add_argument('--clowder_pass',
                        default=os.getenv('CLOWDER_PASS', ''),
                        help='clowder password to use when creating new datasets')

    parser.add_argument('--experiment_json_file', nargs='?', dest='experiment_json_file',
                        default=os.getenv('EXPERIMENT_CONFIG', DEFAULT_EXPERIMENT_JSON_FILENAME),
                        help='Default name of experiment configuration file used to' \
                             ' provide additional processing information')

class TerrarefExtractor(Extractor):

    def __init__(self):

        super(TerrarefExtractor, self).__init__()

        add_arguments(self.parser)
        add_sensor_arguments(self.parser)
        add_influx_arguments(self.parser)

        self.dataset_metadata = None
        self.terraref_metadata = None
        self.experiment_metadata = None
        self.clowderspace = None
        self.debug = None
        self.overwrite = None
        self.clowder_user = None
        self.clowder_pass = None
        self.experiment_json_file = None
        self.logger = None
        self.sensors = None
        self.get_sensor_path = None
        self.influx = None

    def setup(self, base='', site='', sensor=''):

        super(TerrarefExtractor, self).setup()

        self.clowderspace = self.args.clowderspace
        self.debug = self.args.debug
        self.overwrite = self.args.overwrite
        self.clowder_user = self.args.clowder_user
        self.clowder_pass = self.args.clowder_pass
        self.experiment_json_file = self.args.experiment_json_file

        if not base: base = self.args.terraref_base
        if not site: site = self.args.terraref_site
        if not sensor: sensor = self.args.sensor

        #log_config["handlers"]["logstash"]["message_type"] = ("terraref_"+sensor).replace(" ", "_").lower()
        logging.getLogger('pyclowder').setLevel(self.args.debug)
        logging.getLogger('__main__').setLevel(self.args.debug)
        self.logger = logging.getLogger("extractor")

        self.sensors = Sensors(base=base, station=site, sensor=sensor)
        self.get_sensor_path = self.sensors.get_sensor_path

        self.influx = Influx(self.args.influx_host, self.args.influx_port,
                             self.args.influx_db, self.args.influx_user,
                             self.args.influx_pass)

    @property
    def default_epsg(self):
        """Returns the default EPSG code that utilities expect
        """
        return 4326

    @property
    def sensor_name(self):
        """Returns the sensor name the instance is configured with. Returns None if
           sensor information is not available
        """
        if self.sensors:
            return self.sensors.sensor
        return None

    @property
    def config_file_name(self):
        """Returns the name of the expected configuration file
        """
        # pylint: disable=line-too-long
        return DEFAULT_EXPERIMENT_JSON_FILENAME if not self.experiment_json_file \
                                                else self.experiment_json_file
        # pylint: enable=line-too-long

    @property
    def date_format_regex(self):
        """Returns array of regex expressions for different date formats
        """
        # We lead with the best formatting to use, add on the rest
        return [r'(\d{4}(/|-){1}\d{1,2}(/|-){1}\d{1,2})',
                r'(\d{1,2}(/|-){1}\d{1,2}(/|-){1}\d{4})'
               ]

    @property
    def terraref_timestamp_format_regex(self):
        """Returns array of regex expressions for different timestamp formats
        """
        return [r'(\d{4}(/|-){1}\d{1,2}(/|-){1}\d{1,2}__\d{2}-\d{2}-\d{2}[\-{1}\d{3}]*)'
               ]

    @property
    def iso_timestamp_format_regex(self):
        """Returns array of regex expressions for different timestamp formats
        """
        return [r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})[+-](\d{2})\:(\d{2})',
                r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})(.\d+)[+-](\d{2})\:(\d{2})',
                r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})(.\d+)',
                r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})',
                r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})Z'
               ]

    @property
    def dataset_metadata_file_ending(self):
        """ Returns the ending string of a dataset metadata JSON file name
        """
        return '_dataset_metadata.json'

    @property
    def file_infodata_file_ending(self):
        """ Returns the ending string of a file's info JSON file name
        """
        return '_info.json'

    @property
    def get_terraref_metadata(self):
        """Returns any loaded terraref metadata
        """
        return self.terraref_metadata

    @property
    def overwrite_ok(self):
        """Returns whether or not it's OK to overwrite data"""
        if not self.experiment_metadata is None:
            overwrite = __internal__.case_insensitive_find(self.experiment_metadata, 'overwrite')
            if not overwrite is None:
                return overwrite

        return self.overwrite

    def start_check(self, resource):
        """Standard format for extractor logs on check_message."""
        self.logger.info("[%s] %s - Checking message." % (resource['id'], resource['name']))


    def start_message(self, resource):
        self.logger.info("[%s] %s - Processing message." % (resource['id'], resource['name']))
        self.starttime = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self.created = 0
        self.bytes = 0


    def end_message(self, resource):
        self.logger.info("[%s] %s - Done." % (resource['id'], resource['name']))
        endtime = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self.influx.log(self.extractor_info['name'],
                        self.starttime, endtime,
                        self.created, self.bytes)


    def log_info(self, resource, msg):
        """Standard format for extractor logs regarding progress."""
        self.logger.info("[%s] %s - %s" % (resource['id'], resource['name'], msg))


    def log_error(self, resource, msg):
        """Standard format for extractor logs regarding errors/failures."""
        self.logger.error("[%s] %s - %s" % (resource['id'], resource['name'], msg))


    def log_skip(self, resource, msg):
        """Standard format for extractor logs regarding skipped extractions."""
        self.logger.info("[%s] %s - SKIP: %s" % (resource['id'], resource['name'], msg))

    def process_message(self, connector, host, secret_key, resource, parameters):
        """Preliminary handling of a message
        Keyword arguments:
            connector(obj): the message queue connector instance
            host(str): the URI of the host making the connection
            secret_key(str): used with the host API
            resource(dict): dictionary containing the resources associated with the request
            parameters(json): json object of the triggering message contents
        Notes:
            Loads dataset metadata if it's available. Looks for terraref metadata in the dataset
            metadata and stores a reference to that, if available. Looks for an experiment
            configuration file and loads that, if found.
        """
        # Setup to default value
        self.dataset_metadata = None
        self.terraref_metadata = None
        self.experiment_metadata = None

        try:
            # Find the meta data for the dataset and other files of interest
            dataset_file = None
            experiment_file = None
            for onefile in resource['local_paths']:
                if onefile.endswith(self.dataset_metadata_file_ending):
                    dataset_file = onefile
                elif os.path.basename(onefile) == self.config_file_name:
                    experiment_file = onefile
                if not dataset_file is None and not experiment_file is None:
                    break

            # If we don't have dataset metadata already, download it
            dataset_md = None
            if dataset_file is None:
                dataset_id = None
                if 'type' in resource:
                    if resource['type'] == 'dataset':
                        dataset_id = resource['id']
                    elif resource['type'] == 'file' and 'parent' in resource:
                        if 'type' in resource['parent'] and resource['parent']['type'] == 'dataset':
                            dataset_id = resource['parent']['id'] if 'id' in resource['parent'] \
                                                                                    else dataset_id
                if not dataset_id is None:
                    dataset_md = download_dataset_metadata(connector, host, secret_key, dataset_id)
            else:
                # Load the dataset metadata from disk
                dataset_md = load_json_file(dataset_file)

            # If we have terraref metadata then store it and dataset metadata for later use
            if not dataset_md is None:
                terraref_md = get_terraref_metadata(dataset_md)
                if terraref_md:
                    md_len = len(terraref_md)
                    if md_len > 0:
                        self.terraref_metadata = terraref_md

                md_len = len(dataset_md)
                if md_len > 0:
                    self.dataset_metadata = dataset_md
                    self.experiment_metadata = pipeline_get_metadata(dataset_md)

            # Now we load any experiment configuration file
            if not experiment_file is None:
                experiment_md = load_yaml_file(experiment_file)
                if experiment_md:
                    md_len = len(experiment_md)
                    if md_len > 0:
                        cur_experiment_md = self.experiment_metadata
                        new_experiment_md = pipeline_get_metadata(experiment_md)

                        if cur_experiment_md and new_experiment_md:
                            # The contents of the file override the contents of any stored metadata
                            self.experiment_metadata = __internal__.merge_experiment_json(cur_experiment_md,
                                                                                          new_experiment_md)
                        elif cur_experiment_md:
                            self.experiment_metadata = cur_experiment_md
                        else:
                            self.experiment_metadata = new_experiment_md

        # pylint: disable=broad-except
        except Exception as ex:
            self.log_error(resource, "Exception caught while loading dataset metadata: " + str(ex))

    def setup_overrides(self, host, secret_key, resource):
        """Sets up any overrides that may be been specified
        Keyword arguments:
            host(str): the URI of the host making the connection
            secret_key(str): used with the host API
            resource(dict): dictionary containing the resources associated with the request
        Return:
            A function that can be called to restore any modified class instance variables. None
            is returned if a problem ocurrs and the class variables are unchanged
        """
        # Get the best username, password, and space
        old_un, old_pw, old_space = (self.clowder_user, self.clowder_pass, self.clowderspace)
        self.clowder_user, self.clowder_pass, self.clowderspace = \
                                                    self.get_clowder_context(host, secret_key)

        # Ensure that the clowder information is valid
        if not confirm_clowder_info(host, secret_key, self.clowderspace, self.clowder_user,
                                    self.clowder_pass):
            self.log_error(resource, "Clowder configuration is invalid. Not processing " +\
                                     "request")
            self.clowder_user, self.clowder_pass, self.clowderspace = (old_un, old_pw, old_space)
            return None

        # Check for a configured sitename
        old_station = self.sensors.station
        added_station = False
        sitename = None
        if self.experiment_metadata:
            sitename = self.find_sitename()
            if sitename:
                sitename = __internal__.prep_string_for_filename(sitename)
                self.sensors.station = sitename
                if not sitename in STATIONS:
                    added_station = True
                    STATIONS[sitename] = copy.deepcopy(STATIONS['ua-mac'])

        # Change the base path of files to include the user by tweaking the sensor's value
        _, new_base = self.get_username_with_base_path(host, secret_key, resource['id'],
                                                       self.sensors.base)
        sensor_old_base = self.sensors.base
        self.sensors.base = new_base

        def restore_func():
            """Internal function for restoring class variables that's returned to caller
            """
            try:
                self.clowder_user, self.clowder_pass, self.clowderspace = old_un, old_pw, old_space
                self.sensors.base = sensor_old_base
                self.sensors.station = old_station
                if added_station and sitename and sitename in STATIONS:
                    del STATIONS[sitename]
            except Exception as ex:
                logging.warning("Restoring Extractor class variables failed: " + str(ex))

        return restore_func

    # pylint: disable=too-many-nested-blocks
    def extract_datestamp(self, date_string):
        """Extracts the datestamp from a string. The parts of a date can be separated by
           single hyphens or slashes ('-' or '/') and no white space.
        Keyword arguments:
            date_string(str): string to lookup a datestamp in. The first found datestamp is
            returned.
        Returns:
            The extracted datestamp as YYYY-MM-DD. Returns None if a date isn't found
        Notes:
            This function only cares if the timestamp looks correct. It doesn't try to figure
            out if year, month, and day have correct values. The found date string may be
            reformatted to match the expected return.
        """

        # Check the string
        if not __internal__.is_string(date_string):
            return None

        date_string_len = len(date_string)
        if date_string_len <= 0:
            return None

        # Find a date
        for part in date_string.split(' - '):
            for form in self.date_format_regex:
                res = re.search(form, part)
                if res:
                    date = res.group(0).replace('/', '-')
                    # Check for hyphen in first 4 characters to see if we need to move things
                    if not '-' in date[:4]:
                        return date
                    else:
                        split_date = date.split('-')
                        if len(split_date) == 3:
                            return split_date[2] + "-" + split_date[1] + "-" + split_date[0]

        return None

    def extract_timestamp(self, time_string):
        """Extracts a timestamp from the string and returns it
        Args:
            time_string(str): the string in which to look for a timestamp
        Return:
            The found timestamp or None if one isn't found.
        Notes:
            This function will look for both a ISO 8601 long format timestamp as weill as a
            legacy TERRA REF timestamp. If only one is found, that is returned. If both are
            found, the more detailed timestamp is returned. For example, if the TERRA REF
            timestamp only has a date but the ISO timestamp has time as well as date, the ISO
            timestamp will be returned.
        """
        tr_ts, iso_ts = None, None

        # define helper function for parsing the string using regex
        def search_regex(self, time_string, regex_exprs):
            """ Hidden function used for searching a string using an array of regex expressions
            """
            ret_str = None
            ts_parts = time_string.split(" - ")

            # Look for the terra ref timestamp
            for one_part in ts_parts:
                for form in regex_exprs:
                    res = re.search(form, one_part)
                    if res:
                        ret_str = res.group(0)
                        break
                if not ret_str is None:
                    break
            if ret_str is None:
                ret_str = self.extract_datestamp(time_string)
            return ret_str

        # Look for the timestamps
        tr_ts = search_regex(self, time_string, self.terraref_timestamp_format_regex)
        iso_ts = search_regex(self, time_string, self.iso_timestamp_format_regex)

        # If we have none, or one timestamp, return what we have
        if tr_ts is None:
            return iso_ts
        elif iso_ts is None:
            return tr_ts

        # Return the full terraref timestamp if we have it. Otherwise check
        # if the ISO timestamp is more complete and return the most complete
        # timestamp

        # The double undewrscore seperates date from time in TERRA REF
        if '__' in tr_ts:
            return tr_ts

        # If the ISO timestamp contains colons it contains a time and will be better than the
        # date-only TERRA REF string
        return iso_ts if ':' in iso_ts else tr_ts

    def find_datestamp(self, text=None):
        """Returns a date stamp
        Keyword arguments:
            text(str): optional text string to search for a date stamp
        Return:
            A date stamp in the format of YYYY-MM-DD. No checks are made to determine if the
            returned string is a valid date.
        Notes:
            The following places are searched for a valid date; the first date found is the one
            that's returned: the text parameter, the name of the dataset associated with the
            current message being processed, the JSON configuration file as defined by the
            config_file_name() property, the current GMT date.
        """
        datestamp = None

        if not text is None:
            datestamp = self.extract_datestamp(text)

        if datestamp is None and not self.dataset_metadata is None:
            if 'name' in self.dataset_metadata:
                datestamp = self.extract_datestamp(self.dataset_metadata['name'])

        if datestamp is None and not self.experiment_metadata is None:
            datestamp = __internal__.case_insensitive_find(self.experiment_metadata, 'observationTimeStamp')

        if datestamp is None:
            datestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d')

        return datestamp

    def find_timestamp(self, text=None):    # pylint: disable=no-self-use
        """Returns a time stamp consisting of a date and time
        Keyword arguments:
            text(str): optional text string to search for a time stamp
        Return:
            A time stamp in the format of YYYY-MM-DDThh:mm:ss[.sss]|Z|<offset>] or
            YYYY-MM-DD__hh-mm-ss_<id>. No checks are made to determine if the returned string is
            a valid timestamp.
        Notes:
            The following places are searched for a valid timestamp; the first timestamp found is
            the one that's returned: the text parameter, the name of the dataset associated with
            the current message being processed, the JSON configuration file as defined by the
            config_file_name() property, the current GMT timestamp in ISO format.
        """
        timestamp = None

        if not text is None:
            timestamp = self.extract_timestamp(text)

        if timestamp is None and not self.dataset_metadata is None:
            if 'name' in self.dataset_metadata:
                timestamp = self.extract_timestamp(self.dataset_metadata['name'])

        if timestamp is None and not self.experiment_metadata is None:
            timestamp = __internal__.case_insensitive_find(self.experiment_metadata, 'observationTimeStamp')

        if timestamp is None:
            timestamp = datetime.datetime.utcnow().isoformat()

        return timestamp

    def find_sitename(self):
        """Peroforms a shallow search for a sitename in the experiment metadata
        Return:
            Returns a found sitename or None
        """
        if self.experiment_metadata:
            return __internal__.case_insensitive_find(self.experiment_metadata, 'site')

        return None

    def get_username_with_base_path(self, host, key, dataset_id, base_path=None):
        """Looks up the name of the user associated with the dataset. If unable to find
           the user's name from the dataset, the clowder_user variable is used instead.
           If not able to find a valid user name, the string 'unknown' is returned.

        Keyword arguments:
            host(str): the partial URI of the API path including protocol ('/api' portion and
                       after is not needed); assumes a terminating '/'
            key(str): access key for API use
            dataset_id(str): the id of the dataset belonging to the user to lookup
            base_path(str): optional starting path which will have the user name appended
        Return:
            A list of user name and modified base_path.
            The user name as defined in get_dataset_username(), or the specified clowder user
            name, or, finally, the string 'unknown'. Underscores replace whitespace, and invalid
            characters are changed to periods ('.').
            The base_path with the user name appended to it, or None if base_path is None
        """
        try:
            username = get_dataset_username(host, key, dataset_id)
        except Exception:       # pylint: disable=broad-except
            username = None

        # If we don't have a name, see if a user name was specified at runtime
        if (username is None) and (not self.clowder_user is None):
            username = self.clowder_user.strip()
            un_len = len(username)
            if un_len <= 0:
                username = None

        # If we have an experiment configuation, look for a name in there
        if not self.experiment_metadata is None:
            extractors = __internal__.case_insensitive_find(self.experiment_metadata, 'extractors')
            if extractors:
                ex_username = __internal__.case_insensitive_find(extractors, 'firstName')
                last_name = __internal__.case_insensitive_find(extractors, 'lastName')
                if last_name:
                    ex_username = ex_username + ('' if ex_username is None else ' ') + last_name
                if not ex_username is None:
                    username = ex_username

        # Clean up the string
        if not username is None:
            # pylint: disable=line-too-long
            username = __internal__.prep_string_for_filename(username)
            # pylint: enable=line-too-long
        else:
            username = 'unknown'

        # Build up the path if the caller desired that
        new_base_path = None
        if not base_path is None:
            new_base_path = os.path.join(base_path, username)
            new_base_path = new_base_path.rstrip('/')

        return (username, new_base_path)

    def get_season_and_experiment(self, timestamp, sensor):
        """Retrieves the season and experiment from either BETYdb for TERRA REF projects,
           or the experiment configuration file.

        Keyword arguments:
            timestamp(str): The timestamp associated with the season and experiement to return
            sensor(str): Name of the sensor used to retrieve data from BETYdb

        Return:
            Returns a list of the found season name, experiment name, and experiment metadata.
            The returned season name can contain None, 'Unknown Season', of the retrieved season's
            name.
            The returned experiment name can contain None, 'Unknown Experiment', or the
            experiment's name.
            Receiving 'Unknown Season' or 'Unknown Experiment' indicates the lookup failed.
            Experiment metadata is None if it wasn't used during the lookup. Otherwise the
            loaded metadata is returned as the third item in the list (replacing None).
        Notes:
            The TERRA REF lookup is skipped if either, or both, of the timestamp or sensor
            parameters are None, or the terraref metadata was not found for the current message
            being processed.
            The experiment configuration file lookup for season and experiment values is only done
            if the TERRA REF lookup isn't performed.
        """
        season_name, experiment_name, experiment_md = \
                                                    ('Unknown Season', 'Unknown Experiment', None)

        # We first see if the caller is asking for TERRA REF metadata
        if not self.terraref_metadata is None and not timestamp is None and not sensor is None:
            season_name, experiment_name, experiment_md = \
                            get_season_and_experiment(timestamp, sensor, self.terraref_metadata)

        # Look up our fields in the experiment metadata
        if (season_name == 'Unknown Season' or experiment_name == 'Unknown Experiment') and \
                                                            (not self.experiment_metadata is None):
            if season_name == 'Unknown Season':
                season_name = __internal__.case_insensitive_find(self.experiment_metadata, 'season', season_name)

            if experiment_name == 'Unknown Experiment':
                experiment_name = __internal__.case_insensitive_find(self.experiment_metadata, 'studyName', experiment_name)

        return (season_name, experiment_name, experiment_md)

    def get_clowder_context(self, host, key):
        """Returns the best known username, password, and clowder space ID. By default the values
           as defined in the current running environment are returned. If an experiment
           configuration file was loaded, and it has clowder configuration, the values found will
           be returned. No checks are made on the validity of the returned values.

        Keyword arguments:
            host(str): the partial URI of the API path including protocol ('/api' portion and
                       after is not needed); assumes a terminating '/'
            key(str): access key for API use

        Returns:
            A list of username, password, and clowder space ID
        """
        # Set the default return values
        ret_username, ret_password, ret_space = \
                                        (self.clowder_user, self.clowder_pass, self.clowderspace)

        # Check for overrides
        if not self.experiment_metadata is None:
            clowder_md = __internal__.case_insensitive_find(self.experiment_metadata, 'clowder')
            if clowder_md:
                ret_username = __internal__.case_insensitive_find(clowder_md, 'username', ret_username)
                ret_password = __internal__.case_insensitive_find(clowder_md, 'password', ret_password)
                cur_space = __internal__.case_insensitive_find(clowder_md, 'space')

                # Try to look up the space by name, otherwise assume we have an ID
                if cur_space:
                    url = "%sapi/spaces?key=%s&title=%s&exact=true" % (host, key, cur_space)
                    result = requests.get(url)
                    result.raise_for_status()

                    if not len(result.json()) == 0:
                        ret_space = result.json()[0]['id']
                    else:
                        ret_space = cur_space

        return (ret_username, ret_password, ret_space)

    def get_file_filters(self):
        """Gets file filters from the experiment metadata
        Returns:
            Returns the list of file filters if they are found, or None
        """
        file_filters = None
        if self.get_terraref_metadata is None and self.experiment_metadata:
            if 'extractors' in self.experiment_metadata:
                extractor_json = self.experiment_metadata['extractors']
                if self.sensor_name in extractor_json:
                    if 'filters' in extractor_json[self.sensor_name]:
                        file_filters = extractor_json[self.sensor_name]['filters']
                        if ',' in file_filters:
                            file_filters = file_filters.split(',')
                        elif file_filters:
                            file_filters = [file_filters]
        return file_filters


# BASIC UTILS -------------------------------------
def timestamp_to_terraref(timestamp):
    """Converts a timestamp to TERRA REF format
    Args:
        timestamp(str): the ISO timestamp to convert
    Returns:
        The ISO string reformatted to match the TERRA REF time & date stamp. Returns the original
        string if time units aren't specified
    Note:
        Does a simple replacement of the 'T' to '__' and stripping the fraction of seconds and
        timezone information from the string.
    """
    return_ts = timestamp

    if 'T' in timestamp:
        regex_mask = r'(\d{4})-(\d{2})-(\d{2})T(\d{2})\:(\d{2})\:(\d{2})'
        res = re.search(regex_mask, timestamp)
        if res:
            return_ts = res.group(0)
            return_ts = return_ts.replace('T', '__').replace(':', '-')

    return return_ts

def build_metadata(clowderhost, extractorinfo, target_id, content, target_type='file', context=None):
    """Construct extractor metadata object ready for submission to a Clowder file/dataset.

        clowderhost -- root URL of Clowder target instance (before /api)
        extractorname -- name of extractor, in extractors usually self.extractor_info['name']
        target_id -- UUID of file or dataset that metadata will be sent to
        content -- actual JSON contents of metadata
        target_type -- type of target resource, 'file' or 'dataset'
        context -- (optional) list of JSON-LD contexts
    """
    if context is None:
        context = ["https://clowder.ncsa.illinois.edu/contexts/metadata.jsonld"]

    content['extractor_version'] = extractorinfo['version']

    md = {
        # TODO: Generate JSON-LD context for additional fields
        "@context": context,
        "content": content,
        "agent": {
            "@type": "cat:extractor",
            "extractor_id": clowderhost + ("" if clowderhost.endswith("/") else "/") + "api/extractors/"+ extractorinfo['name'],
            "version": extractorinfo['version'],
            "name": extractorinfo['name']
        }
    }

    if target_type == 'dataset':
        md['dataset_id'] = target_id
    else:
        md['file_id'] = target_id

    return md


def is_latest_file(resource):
    """Check whether the extractor-triggering file is the latest file in the dataset.

    This simple check should be used in dataset extractors to avoid collisions between 2+ instances of the same
    extractor trying to process the same dataset simultaneously by triggering off of 2 different uploaded files.

    Note that in the resource dictionary, "triggering_file" is the file that triggered the extraction (i.e. latest file
    at the time of message generation), not necessarily the newest file in the dataset.
    """
    trig = None
    if 'triggering_file' in resource:
        trig = resource['triggering_file']
    elif 'latest_file' in resource:
        trig = resource['latest_file']

    if trig:
        latest_file = ""
        latest_time = "Sun Jan 01 00:00:01 CDT 1920"

        for f in resource['files']:
            try:
                create_time = datetime.datetime.strptime(f['date-created'], "%a %b %d %H:%M:%S %Z %Y")
                latest_dt = datetime.datetime.strptime(latest_time, "%a %b %d %H:%M:%S %Z %Y")

                if f['filename'] == trig:
                    trig_dt = create_time

                if create_time > latest_dt:
                    latest_time = f['date-created']
                    latest_file = f['filename']
            except:
                return True

        if latest_file == trig or latest_time == trig_dt:
            return True
        else:
            return False
    else:
        # If unable to determine triggering file, return True
        return True


def contains_required_files(resource, required_list):
    """Iterate through files in resource and check if all of required list is found."""
    for req in required_list:
        found_req = False
        for f in resource['files']:
            if f['filename'].endswith(req):
                found_req = True
        if not found_req:
            return False
    return True


def load_json_file(filepath):
    """Load contents of a .json file on disk into a JSON object.
    """
    try:
        with open(filepath, 'r') as jsonfile:
            return json.load(jsonfile)
    except:
        logging.error('could not load .json file %s' % filepath)
        return None


def load_yaml_file(filepath):
    """Load contents of a YAML file on disk into a Python object.

    Keyword arguments:
        filepath(str): the path to the YAML file to load
    Return:
        The python object representing the YAML file contents. None is returned if the file
        couldn't be loaded.
    """
    try:
        with open(filepath, 'r') as yamlfile:
            return yaml.safe_load(yamlfile)
    except:
        logging.error('could not load YAML file %s' % filepath)
        return None


def file_exists(filepath, max_age_mins=3):
    """Return True if a file already exists on disk.

    If the file is zero bytes in size, return False if the file is more than
    max_age minutes old.
    """

    if os.path.exists(filepath):
        if os.path.getsize(filepath) > 0:
            return True
        else:
            age_seconds = time.time() - os.path.getmtime(filepath)
            return age_seconds < (max_age_mins*60)
    else:
        return False

# CLOWDER UTILS -------------------------------------
# TODO: Remove redundant ones of these once PyClowder2 supports user/password
def build_dataset_hierarchy(host, secret_key, clowder_user, clowder_pass, root_space,
                            season, experiment, root_coll_name, year='', month='', date='', leaf_ds_name=''):
    """This will build collections if needed in parent space.

        Typical hierarchy:
        MAIN LEVEL 1 DATA SPACE IN CLOWDER
        - Season ("Season 6")
        - Experiment ("Sorghum BAP")
        - Root collection for sensor ("stereoRGB geotiffs")
            - Year collection ("stereoRGB geotiffs - 2017")
                - Month collection ("stereoRGB geotiffs - 2017-01")
                    - Date collection ("stereoRGB geotiffs - 2017-01-01")
                        - Dataset ("stereoRGB geotiffs - 2017-01-01__01-02-03-456")

        Omitting year, month or date will result in dataset being added to next level up.
    """
    if season:
        season_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, season,
                                                  parent_space=root_space)

        experiment_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, experiment,
                                                      season_collect, parent_space=root_space)

        sensor_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, root_coll_name,
                                                  experiment_collect, parent_space=root_space)
    elif experiment:
        experiment_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, experiment,
                                                      parent_space=root_space)

        sensor_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, root_coll_name,
                                                  experiment_collect, parent_space=root_space)
    else:
        sensor_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, root_coll_name,
                                                  parent_space=root_space)

    if year:
        # Create year-level collection
        year_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass,
                                                "%s - %s" % (root_coll_name, year),
                                                sensor_collect, parent_space=root_space)
        #verify_collection_in_space(host, secret_key, year_collect, root_space)
        if month:
            # Create month-level collection
            month_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass,
                                                     "%s - %s-%s" % (root_coll_name, year, month),
                                                     year_collect, parent_space=root_space)
            #verify_collection_in_space(host, secret_key, month_collect, root_space)
            if date:
                targ_collect = get_collection_or_create(host, secret_key, clowder_user, clowder_pass,
                                                        "%s - %s-%s-%s" % (root_coll_name, year, month, date),
                                                        month_collect, parent_space=root_space)
                #verify_collection_in_space(host, secret_key, targ_collect, root_space)
            else:
                targ_collect = month_collect
        else:
            targ_collect = year_collect
    else:
        targ_collect = sensor_collect

    target_dsid = get_dataset_or_create(host, secret_key, clowder_user, clowder_pass, leaf_ds_name,
                                        targ_collect, root_space)
    #verify_dataset_in_space(host, secret_key, target_dsid, root_space)
    return target_dsid


def build_dataset_hierarchy_crawl(host, secret_key, clowder_user, clowder_pass, root_space,
                                  season=None, experiment=None, sensor=None, year=None, month=None, date=None, leaf_ds_name=None):
    """This will build collections if needed in parent space.

        Typical hierarchy:
        MAIN LEVEL 1 DATA SPACE IN CLOWDER
        - Season ("Season 6")
        - Experiment ("Sorghum BAP")
        - Root collection for sensor ("stereoRGB geotiffs")
            - Year collection ("stereoRGB geotiffs - 2017")
                - Month collection ("stereoRGB geotiffs - 2017-01")
                    - Date collection ("stereoRGB geotiffs - 2017-01-01")
                        - Dataset ("stereoRGB geotiffs - 2017-01-01__01-02-03-456")

        Omitting year, month or date will result in dataset being added to next level up.

        Start at the root collection and check children until we get to the final one.
    """
    if season and experiment and sensor:
        season_c = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, season, parent_space=root_space)
        experiment_c = ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, root_space, season_c, experiment)
        sensor_c = ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, root_space, experiment_c, sensor)
    elif sensor:
        sensor_c = get_collection_or_create(host, secret_key, clowder_user, clowder_pass, sensor, parent_space=root_space)
    else:
        sensor_c = None

    if year:
        year_c_name = "%s - %s" % (sensor, year)
        year_c = ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, root_space, sensor_c, year_c_name)

        if month:
            month_c_name = "%s - %s-%s" % (sensor, year, month)
            month_c = ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, root_space, year_c, month_c_name)

            if date:
                date_c_name = "%s - %s-%s-%s" % (sensor, year, month, date)
                targ_c = ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, root_space, month_c, date_c_name)

            else:
                targ_c = month_c
        else:
            targ_c = year_c
    else:
        targ_c = sensor_c

    target_dsid = get_dataset_or_create(host, secret_key, clowder_user, clowder_pass, leaf_ds_name, targ_c, root_space)
    return target_dsid


def get_collection_or_create(host, secret_key, clowder_user, clowder_pass, cname, parent_colln=None, parent_space=None):
    # Fetch dataset from Clowder by name, or create it if not found
    url = "%sapi/collections?key=%s&title=%s&exact=true" % (host, secret_key, cname)
    result = requests.get(url)
    result.raise_for_status()

    if len(result.json()) == 0:
        return create_empty_collection(host, clowder_user, clowder_pass, cname, "", parent_colln, parent_space)
    else:
        coll_id = result.json()[0]['id']
        if parent_colln:
            add_collection_to_collection(host, secret_key, parent_colln, coll_id)
        if parent_space:
            add_collection_to_space(host, secret_key, coll_id, parent_space)
        return coll_id

def get_child_collections(host, secret_key, collection_id):
    url = "%sapi/collections/%s/getChildCollections?key=%s" % (host, collection_id, secret_key)
    result = requests.get(url)
    result.raise_for_status()

    return result.json()

def create_empty_collection(host, clowder_user, clowder_pass, collectionname, description, parentid=None, spaceid=None):
    """Create a new collection in Clowder.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    clowder_user -- the username to login to clowder
    clowder_pass -- the password associated with the username
    collectionname -- name of new dataset to create
    description -- description of new dataset
    parentid -- id of parent collection
    spaceid -- id of the space to add dataset to
    """

    logger = logging.getLogger(__name__)

    if parentid:
        if (spaceid):
            url = '%sapi/collections/newCollectionWithParent' % host
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": collectionname, "description": description,
                                                    "parentId": parentid, "space": spaceid}),
                                   auth=(clowder_user, clowder_pass))
        else:
            url = '%sapi/collections/newCollectionWithParent' % host
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": collectionname, "description": description,
                                                    "parentId": parentid}),
                                   auth=(clowder_user, clowder_pass))
    else:
        if (spaceid):
            url = '%sapi/collections' % host
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": collectionname, "description": description,
                                                    "space": spaceid}),
                                   auth=(clowder_user, clowder_pass))
        else:
            url = '%sapi/collections' % host
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": collectionname, "description": description}),
                                   auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    collectionid = result.json()['id']
    logger.debug("collection id = [%s]", collectionid)

    return collectionid

def get_dataset_or_create(host, secret_key, clowder_user, clowder_pass, dsname, parent_colln=None, parent_space=None):
    # Fetch dataset from Clowder by name, or create it if not found
    ds_id = get_datasetid_by_name(host, secret_key, dsname)

    if not ds_id:
        return create_empty_dataset(host, clowder_user, clowder_pass, dsname, "",
                                    parent_colln, parent_space)
    else:
        if parent_colln:
            add_dataset_to_collection(host, secret_key, ds_id, parent_colln)
        if parent_space:
            add_dataset_to_space(host, secret_key, ds_id, parent_space)
        return ds_id

def create_empty_dataset(host, clowder_user, clowder_pass, datasetname, description, parentid=None, spaceid=None):
    """Create a new dataset in Clowder.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    datasetname -- name of new dataset to create
    description -- description of new dataset
    parentid -- id of parent collection
    spaceid -- id of the space to add dataset to
    """

    logger = logging.getLogger(__name__)

    url = '%sapi/datasets/createempty' % host

    if parentid:
        if spaceid:
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": datasetname, "description": description,
                                                    "collection": [parentid], "space": [spaceid]}),
                                   auth=(clowder_user, clowder_pass))
        else:
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": datasetname, "description": description,
                                                    "collection": [parentid]}),
                                   auth=(clowder_user, clowder_pass))
    else:
        if spaceid:
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": datasetname, "description": description,
                                                    "space": [spaceid]}),
                                   auth=(clowder_user, clowder_pass))
        else:
            result = requests.post(url, headers={"Content-Type": "application/json"},
                                   data=json.dumps({"name": datasetname, "description": description}),
                                   auth=(clowder_user, clowder_pass))

    result.raise_for_status()

    datasetid = result.json()['id']
    logger.debug("dataset id = [%s]", datasetid)

    return datasetid

def upload_to_dataset(connector, host, clowder_user, clowder_pass, datasetid, filepath):
    """Upload file to existing Clowder dataset.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    datasetid -- the dataset that the file should be associated with
    filepath -- path to file
    check_duplicate -- check if filename already exists in dataset and skip upload if so
    """

    logger = logging.getLogger(__name__)

    for source_path in connector.mounted_paths:
        if filepath.startswith(connector.mounted_paths[source_path]):
            return _upload_to_dataset_local(connector, host, clowder_user, clowder_pass, datasetid, filepath)

    url = '%sapi/uploadToDataset/%s' % (host, datasetid)

    if os.path.exists(filepath):
        result = connector.post(url, files={"File": open(filepath, 'rb')},
                                auth=(clowder_user, clowder_pass))

        uploadedfileid = result.json()['id']
        logger.debug("uploaded file id = [%s]", uploadedfileid)

        return uploadedfileid
    else:
        logger.error("unable to upload file %s (not found)", filepath)
        return None

def _upload_to_dataset_local(connector, host, clowder_user, clowder_pass, datasetid, filepath):
    """Upload file POINTER to existing Clowder dataset. Does not copy actual file bytes.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    datasetid -- the dataset that the file should be associated with
    filepath -- path to file
    """

    logger = logging.getLogger(__name__)
    url = '%sapi/uploadToDataset/%s' % (host, datasetid)

    if os.path.exists(filepath):
        # Replace local path with remote path before uploading
        for source_path in connector.mounted_paths:
            if filepath.startswith(connector.mounted_paths[source_path]):
                filepath = filepath.replace(connector.mounted_paths[source_path],
                                            source_path)
                break

        (content, header) = encode_multipart_formdata([
            ("file", '{"path":"%s"}' % filepath)
        ])
        result = connector.post(url, data=content, headers={'Content-Type': header},
                                auth=(clowder_user, clowder_pass))

        uploadedfileid = result.json()['id']
        logger.debug("uploaded file id = [%s]", uploadedfileid)

        return uploadedfileid
    else:
        logger.error("unable to upload local file %s (not found)", filepath)
        return None

def get_child_collections(host, secret_key, collectionid):
    """Get list of child collections in collection by UUID.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    collectionid -- the collection to get children of
    """

    if collectionid:
        url = "%sapi/collections/%s/getChildCollections?key=%s" % (host, collectionid, secret_key)

        result = requests.get(url)
        result.raise_for_status()

        return json.loads(result.text)
    else:
        return []

def get_datasets(host, clowder_user, clowder_pass, collectionid):
    """Get list of datasets in collection by UUID.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    datasetid -- the collection to get datasets of
    """

    url = "%sapi/collections/%s/datasets" % (host, collectionid)

    result = requests.get(url, auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    return json.loads(result.text)

def delete_dataset(host, clowder_user, clowder_pass, datasetid):
    url = "%sapi/datasets/%s" % (host, datasetid)

    result = requests.delete(url, auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    return json.loads(result.text)

def delete_dataset_metadata(host, clowder_user, clowder_pass, datasetid):
    url = "%sapi/datasets/%s/metadata.jsonld" % (host, datasetid)

    result = requests.delete(url, stream=True, auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    return json.loads(result.text)

def delete_collection(host, clowder_user, clowder_pass, collectionid):
    url = "%sapi/collections/%s" % (host, collectionid)

    result = requests.delete(url, auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    return json.loads(result.text)

def delete_dataset_metadata_in_collection(host, clowder_user, clowder_pass, collectionid, recursive=True):
    dslist = get_datasets(host, clowder_user, clowder_pass, collectionid)

    logging.info("deleting dataset metadata in collection %s" % collectionid)
    for ds in dslist:
        delete_dataset_metadata(host, clowder_user, clowder_pass, ds['id'])
    logging.info("completed %s datasets" % len(dslist))

    if recursive:
        childcolls = get_child_collections(host, clowder_user, clowder_pass, collectionid)
        for coll in childcolls:
            delete_dataset_metadata_in_collection(host, clowder_user, clowder_pass, coll['id'], recursive)

def delete_datasets_in_collection(host, clowder_user, clowder_pass, collectionid, recursive=True, delete_colls=True):
    dslist = get_datasets(host, clowder_user, clowder_pass, collectionid)

    logging.info("deleting datasets in collection %s" % collectionid)
    for ds in dslist:
        delete_dataset(host, clowder_user, clowder_pass, ds['id'])
    logging.info("completed %s datasets" % len(dslist))

    if recursive:
        childcolls = get_child_collections(host, clowder_user, clowder_pass, collectionid)
        for coll in childcolls:
            delete_datasets_in_collection(host, clowder_user, clowder_pass, coll['id'], recursive, delete_colls)

    if delete_colls:
        logging.info("deleting collection %s" % collectionid)
        delete_collection(host, clowder_user, clowder_pass, collectionid)

def get_datasetid_by_name(host, secret_key, dsname):
    """Looks up the ID of a dataset by nanme

    Args:
        host(str): the URI of the host making the connection
        secret_key(str): used with the host API
        dsname(str): the dataset name to look up

    Return:
        Returns the ID of the dataset if it's found. Returns None if the dataset
        isn't found
    """
    url = "%sapi/datasets?key=%s&title=%s&exact=true" % (host, secret_key, dsname)

    try:
        result = requests.get(url)
        result.raise_for_status()

        md = result.json()
        md_len = len(md)
    except Exception as ex:     # pylint: disable=broad-except
        md = None
        md_len = 0
        logging.debug(ex.message)

    if md and md_len > 0 and "id" in md[0]:
        return md[0]["id"]

    return None

def create_empty_space(host, clowder_user, clowder_pass, space_name, description=""):
    """Create a new space in Clowder.

    Keyword arguments:
    connector -- connector information, used to get missing parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    space_name -- name of new space to create
    """

    logger = logging.getLogger(__name__)

    url = '%sapi/spaces' % host
    result = requests.post(url, headers={"Content-Type": "application/json"},
                           data=json.dumps({"name": space_name, "description": description}),
                           auth=(clowder_user, clowder_pass))
    result.raise_for_status()

    spaceid = result.json()['id']
    logger.debug("space id = [%s]", spaceid)

    return spaceid

def get_space_or_create(host, secret_key, clowder_user, clowder_pass, space_name):
    # Fetch dataset from Clowder by name, or create it if not found
    url = "%sapi/spaces?key=%s&title=%s&exact=true" % (host, secret_key, space_name)
    result = requests.get(url)
    result.raise_for_status()

    if len(result.json()) == 0:
        return create_empty_collection(host, clowder_user, clowder_pass, space_name, "")
    else:
        return result.json()[0]['id']

def delete_file(host, secret_key, fileid):
    url = "%sapi/files/%s?key=%s" % (host, fileid, secret_key)
    result = requests.delete(url)
    result.raise_for_status()

def check_file_in_dataset(connector, host, secret_key, dsid, filepath, remove=False, forcepath=False, replacements=None):
    # Replacements = [("L2","L1")]
    # Each tuple is checked replacing first element in filepath with second element for existing
    dest_files = get_file_list(connector, host, secret_key, dsid)

    if replacements and len(replacements) > 0:
        for r in replacements:
            filepath = filepath.replace(r[0], r[1])

    for source_path in connector.mounted_paths:
        if filepath.startswith(connector.mounted_paths[source_path]):
            filepath = filepath.replace(connector.mounted_paths[source_path], source_path)

    filename = os.path.basename(filepath)

    found_file = False
    for f in dest_files:
        if (not forcepath and f['filename'] == filename) or (forcepath and f['filepath'] == filepath):
            if remove:
                delete_file(host, secret_key, f['id'])
            found_file = True

    return found_file

def ensure_collection_in_children(host, secret_key, clowder_user, clowder_pass, parent_space, parent_coll_id, child_name):
    """Check if named collection is among parent's children, and create if not found."""
    child_collections = get_child_collections(host, secret_key, parent_coll_id)
    for c in child_collections:
        if c['name'] == child_name:
            return str(c['id'])

    # If we didn't find it, create it
    return create_empty_collection(host, clowder_user, clowder_pass, child_name, "", parent_coll_id, parent_space)

def add_dataset_to_collection(host, secret_key, dataset_id, collection_id):
    # Didn't find space, so we must associate it now
    url = "%sapi/collections/%s/datasets/%s?key=%s" % (host, collection_id, dataset_id, secret_key)
    result = requests.post(url)
    result.raise_for_status()

def add_dataset_to_space(host, secret_key, dataset_id, space_id):
    # Didn't find space, so we must associate it now
    url = "%sapi/spaces/%s/addDatasetToSpace/%s?key=%s" % (host, space_id, dataset_id, secret_key)
    result = requests.post(url)
    result.raise_for_status()

def add_collection_to_collection(host, secret_key, parent_coll_id, child_coll_id):
    # Didn't find space, so we must associate it now
    url = "%sapi/collections/%s/addSubCollection/%s?key=%s" % (host, parent_coll_id, child_coll_id, secret_key)
    result = requests.post(url)
    result.raise_for_status()

def add_collection_to_space(host, secret_key, collection_id, space_id):
    # Didn't find space, so we must associate it now
    url = "%sapi/spaces/%s/addCollectionToSpace/%s?key=%s" % (host, space_id, collection_id, secret_key)
    result = requests.post(url)
    result.raise_for_status()

def confirm_clowder_info(host, secret_key, space_id, clowder_user=None, clowder_pass=None):
    """Confirms that the information provided is valid in the clowder instance

    Keyword arguments:
        host(str): the partial URI of the API path including protocol ('/api' portion and
                   after is not needed); assumes a terminating '/'
        secret_key(str): access key for API use
        space_id(str): the id of the space to check
        clowder_user(str): the clowder username
        clowder_pass(str): the password associated with the username

    Returns:
        True is returned if the parameters appear to be good. False is returned otherwise
    """
    logger = logging.getLogger(__name__)

    # Check that we have good parameters
    if not secret_key or not space_id:
        logger.error("One or more required parameters is empty")
        return False
    if (not clowder_user and clowder_pass) or (clowder_user and not clowder_pass):
        logger.error("Both a user name and password must be specified or set to None, only one "\
                     "value was was detected")
        return False

    # Now check with clowder
    try:
        # First try to find the user name
        found = find_user_name(host, secret_key, clowder_user)
        if not found:
            logger.info("Clowder user not found by querying users: %s", clowder_user)

        # Try to find the space in Clowder
        url = '%sapi/spaces/%s?key=%s' % (host, space_id, secret_key)
        result = requests.get(url)
        result.raise_for_status()

        ret = result.json()
        found = False
        if ('id' in ret) and (ret['id'] == space_id):
            found = True
        if not found:
            logger.info("Clowder space not found: %s", space_id)
            return False
    # pylint: disable=broad-except
    except Exception as ex:
        logger.error("Exception caught checking clowder information: %s", str(ex))
        return False

    return True

def file_filtered_in(filename, filters):
    """Checks the filename against a list of filters
    Keyword arguments:
        filename(str): the filename to check
        filters(list): list of string parts to check the filename against
    Return:
        True is returned if a filter matches part of the file name. False if no match found
    """
    filtered = False
    if filters:
        for one_filter in filters:
            if one_filter in os.path.basename(filename):
                filtered = True
                break
    return filtered

def _search_for_key(metadata, key_variants):
    """Check for presence of any key variants in metadata. Does basic capitalization check.

        Returns:
        value if found, or None
    """
    val = None
    for variant in key_variants:
        if variant in metadata:
            val = metadata[variant]
        elif variant.capitalize() in metadata:
            val = metadata[variant.capitalize()]

    # If a value was found, try to parse as float
    if val:
        try:
            return float(val.encode("utf-8"))
        except:
            return val
    else:
        return None



"""Provides data security functions
"""

import os
import logging
import math
import sys
import base64
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

# The names of the environment variables associated with encryption
PIPELINE_KEY_NAME = "PIPELINE_KEY"

# Attempt to get the encryption envrionment variables
CRYPT_KEY = os.getenv(PIPELINE_KEY_NAME)

# Lengths of encryption key, iv (initialization vector), and other required sizes
CRYPT_KEY_BYTE_LEN = int((256 / 8)) # 256 bits length
CRYPT_IV_BYTE_LEN = (16)            # We are using AES -> block size is 16
CRYPT_SOURCE_BLOCK_SIZE = (16)      # Reesize the source block to a multiple of this value; zero for no change

def _sized_byte_like(value, size):
    """Make the value byte like and the correct size. If the value is not the correct size
       it will be either padded with spaces or trancated
    Args:
        value(str, bytes, bytearray): The value to make byte-like and correctly sized
        size(int): The correct size of the value
    Return:
        Returns the correctly sized value as bytes
    """
    if isinstance(value, str):
        value_len = len(value)
        new_value = b''
        for idx in range(0, value_len):
            if sys.version_info[0] < 3:
                new_value += bytes(value[idx])
            else:
                new_value += bytes([ord(value[idx])])
    else:
        new_value = value

    value_len = len(new_value)
    while value_len < size:
        new_value += b' '
        value_len += 1
    if value_len > size:
        new_value = new_value[:size]

    return new_value

def _perform_encrypt(plain_text):
    """Encrypts the plain text using environmental variables
    Args:
        plain_text(str): the string to encrypt
    Returns:
        Returns a tuple of the encrypted string, the length of the plain text string, and the IV. If
        a problem occurs then a tuple of (None, 0, None) is returned
    Notes:
        Will log warnings and return None if the PIPELINE_KEY or PIPELINE_IV environment
        variables haven't been set
    """
    global CRYPT_KEY                # pylint: disable=global-statement
    global CRYPT_IV_BYTE_LEN        # pylint: disable=global-statement
    global CRYPT_KEY_BYTE_LEN       # pylint: disable=global-statement
    global CRYPT_SOURCE_BLOCK_SIZE  # pylint: disable=global-statement
    global PIPELINE_KEY_NAME        # pylint: disable=global-statement

    if not CRYPT_KEY:
        logging.warning("Missing %s environment variable needed for securing data", PIPELINE_KEY_NAME)
        return (None, 0, None)

    # Setup the encryption variables
    crypt_iv = os.urandom(CRYPT_IV_BYTE_LEN)
    size_the_str = lambda source, length: source if len(source) == length else source[:length] if len(source) > length \
                            else '{str:{fill}{align}{width}}'.format(str=source, fill=' ', align='<', width=length)

    # The key needs to be byte-like and the right size
    key = _sized_byte_like(CRYPT_KEY, CRYPT_KEY_BYTE_LEN)

    # See if we need to resize the source string
    encrypt_text = plain_text
    original_plain_len = len(plain_text)
    if CRYPT_SOURCE_BLOCK_SIZE:
        plain_block_count = math.trunc(float(original_plain_len + CRYPT_SOURCE_BLOCK_SIZE - 1.0) / \
                                                                                        float(CRYPT_SOURCE_BLOCK_SIZE))
        if plain_block_count * CRYPT_SOURCE_BLOCK_SIZE > original_plain_len:
            encrypt_text = size_the_str(encrypt_text, plain_block_count * CRYPT_SOURCE_BLOCK_SIZE)

    # Perform the encryption
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CBC(crypt_iv), backend=backend)
    encryptor = cipher.encryptor()
    result = encryptor.update(encrypt_text.encode()) + encryptor.finalize()

    return (result, original_plain_len, crypt_iv)

def _perform_decrypt(encrypted, plain_len, crypt_iv):
    """Decrypts an encrypted string.
    Args:
        encrypted(str): the encrypted string
        plain_len(int): the length of the recovered string (in cases where padding ocurred)
        crypt_iv(str): the iv (initialization vector) of the encryption
    Return:
        The un-encrypted string or None if a problem is found
    Note:
        No checks are made to the suitability of the parameters for decryption. For example,
        the iv parameter is not checked to see if it's a valid size
    """
    global CRYPT_KEY   # pylint: disable=global-statement

    if not CRYPT_KEY:
        logging.warning("Missing %s environment variable needed for securing data", PIPELINE_KEY_NAME)
        return None

    # The key and iv needs to be byte-like and the right size
    key = _sized_byte_like(CRYPT_KEY, CRYPT_KEY_BYTE_LEN)
    crypt_iv = _sized_byte_like(crypt_iv, CRYPT_IV_BYTE_LEN)

    # Perform the decryption
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CBC(crypt_iv), backend=backend)
    decryptor = cipher.decryptor()
    plain_text = decryptor.update(encrypted) + decryptor.finalize()

    recovered_len = len(plain_text)
    return plain_text if recovered_len <= plain_len else plain_text[:plain_len]

def _package_encrypted(encrypted, plain_len, crypt_iv):
    """Packages up the parameters into a base64 encoded string
    Args:
        encrypted(bytes, bytearray): an string representing encrypted data
        plain_len(int): the length of the plain text data
        crypt_iv(bytes, bytearray): the iv (initialization vector) value
    Return:
        A base64 string representing the data.
    Note:
        The strings are concatenated as 6 characters each for the length
        of the encrypted parameter and plain_len, followed by the encrypted
        parameter contents, and finally the iv value. The concatenated string
        is then base64 encoded and returned
    """
    if not isinstance(encrypted, (bytes, bytearray)):
        raise RuntimeError("Invalid encrypted text passed into _package_encrypted() - must be byte-like")

    encrypt_len = '{0:06d}'.format(len(encrypted))
    len_str = '{0:06d}'.format(plain_len)

    return base64.b64encode(encrypt_len.encode('utf-8') + len_str.encode('utf-8') + encrypted + \
                            crypt_iv).decode('utf-8')

def _unpackage_encoded(encoded):
    """The encoded string fron _package_encrypted()
    Args:
        encoded(str): the base64 encoded string to take apart
    Return:
        Returns the tuple of encoded string, plain text length, and the iv as a string
    """
    base_str = base64.b64decode(encoded)

    encrypt_len = int(base_str[0:6].decode('utf-8'))
    plain_len = int(base_str[6:12].decode('utf-8'))
    encrypted = base_str[12:encrypt_len+12]
    crypt_iv = base_str[encrypt_len+12:]

    return (encrypted, plain_len, crypt_iv)

def encrypt_pipeline_string(plain_text):
    """Fixes the metadata so that the drone pipeline easily reference it
    Args:
        plain_text(str): the string to secure
    Returns:
        The secured string or None if the string can't be secured
    """
    try:
        encrypted, orig_len, crypt_iv = _perform_encrypt(plain_text)
        if not encrypted is None:
            return _package_encrypted(encrypted, orig_len, crypt_iv)
    except Exception as ex:
        logging.warning("Exception caught while encrypting string: %s", str(ex))

    return None

def decrypt_pipeline_string(crypt_str):
    """Recovers a string that has been secured.
    Args:
        crypt_str(str): the string containing information to recover
    Return:
        The decrypted string when successful and None when not
    """
    return_str = None

    try:
        # Decode the base64 string and get the parts
        encrypted, plain_len, crypt_iv = _unpackage_encoded(crypt_str)

        # Decrypt the string
        plain_str = _perform_decrypt(encrypted, plain_len, crypt_iv)
        plain_str_len = len(plain_str)

        # Make sure we have what is expected
        if plain_str_len == plain_len:
            return_str = plain_str
        else:
            logging.warning("Decrypting string failed - recovered plain text has the wrong size: >> %s <<",
                            str(crypt_str))
    except Exception as ex:
        logging.warning("Exception caught decrypting string: %s", str(ex))

    return return_str
